Package: rapsimng
Type: Package
Title: APSIM Next Generation
Version: 0.6.0
Authors@R: person("Bangyou", "Zheng", 
               email = "zheng.bangyou@gmail.com", 
               role = c("aut", "cre"))
Description: The Agricultural Production Systems sIMulator ('APSIM') is a widely
    used to simulate the agricultural systems for multiple crops. This package 
    is designed to create, modify and run 'apsimx' files in the 'APSIM' Next 
    Generation <https://www.apsim.info/>.
License: MIT + file LICENSE
URL: https://rapsimng.bangyou.me/, https://github.com/byzheng/rapsimng
BugReports: https://github.com/byzheng/rapsimng/issues
Encoding: UTF-8
Depends: R (>= 4.1.0)
Imports: jsonlite, tibble, magrittr, dplyr, purrr, rlang, DBI, RSQLite,
        tidyweather
Suggests: testthat, knitr, rmarkdown, tidyverse, methods
VignetteBuilder: knitr
Config/roxygen2/version: 8.1.0
NeedsCompilation: no
Packaged: 2026-08-24 06:09:38 UTC; zhe00a
Author: Bangyou Zheng [aut, cre]
Maintainer: Bangyou Zheng <zheng.bangyou@gmail.com>
Repository: CRAN
Date/Publication: 2026-08-24 06:30:02 UTC
Built: R 4.5.2; ; 2026-08-24 09:43:24 UTC; unix

# Generated by roxygen2: do not edit by hand

export(append_model)
export(calc_cum_damage_frost_heat)
export(calc_daily_damage_frost_heat)
export(cum_damage_frost)
export(cum_damage_frost_heat)
export(cum_damage_heat)
export(daily_damage_frost)
export(daily_damage_frost_heat)
export(daily_damage_heat)
export(default_frost_heat_params)
export(disable_models)
export(get_cultivar)
export(get_experiment)
export(get_metfile)
export(get_parent)
export(get_pawc)
export(get_simulations)
export(insert_model)
export(insert_models)
export(install_apsimng)
export(keep_simulations)
export(list_report)
export(log_level)
export(minimum_apsimng)
export(new_cultivar)
export(new_model)
export(read_apsimx)
export(read_report)
export(remove_model)
export(replace_model)
export(run_models)
export(search_node)
export(search_path)
export(set_parameter_value)
export(split_apsimx)
export(stage_sens_frost)
export(stage_sens_heat)
export(temp_damage_frost)
export(temp_damage_heat)
export(update_cultivar)
export(with_apsimx)
export(write_apsimx)
importFrom(magrittr,"%>%")
importFrom(rlang,.data)
importFrom(tibble,tibble)

# rapsimng
[![](https://www.r-pkg.org/badges/version/rapsimng?color=green)](https://cran.r-project.org/package=rapsimng)
![R-CMD-check](https://github.com/byzheng/rapsimng/workflows/R-CMD-check/badge.svg) 

[![](http://cranlogs.r-pkg.org/badges/grand-total/rapsimng?color=green)](https://cran.r-project.org/package=rapsimng)
[![](http://cranlogs.r-pkg.org/badges/last-month/rapsimng?color=green)](https://cran.r-project.org/package=rapsimng)
[![](http://cranlogs.r-pkg.org/badges/last-week/rapsimng?color=green)](https://cran.r-project.org/package=rapsimng)


[rapsimng](https://rapsimng.bangyou.me) package is designed to modify and run the *.apsimx simulations using R in [APSIM Next Generation](https://www.apsim.info/).



## Features

* Find elements in the *.apsimx file


## Crop-specific resources

For crop-specific functions and additional resources, please refer to:

* [Wheat](https://wheat.rapsimng.bangyou.me/)
* [Canola](https://canola.rapsimng.bangyou.me/)


## Installation

Install from CRAN.

```r
install.packages('rapsimng')
```


Install the developing version from [Github](https://github.com/byzheng/rapsimng).

```r
remotes::install_github('byzheng/rapsimng')
```


## Read APSIMX file

The `wheat.apsimx` in the validation dataset of APSIM NG  is used as an example. Function `read_apsimx` is used to read `*.apsimx` file through `jsonlite::read_json` and returns as a list.


```r
# Read Wheat.apsimx file with `read_apsimx` which returns a list of json results.
file <- system.file("extdata/wheat.apsimx", package = "rapsimng")
m <- read_apsimx(file)

```



## Search by of APSIM NG

A node in the apsimx file can be found using the path specification in APSIM NG.
```r
potential <- search_path(m,
    path = '[Structure].BranchingRate.PotentialBranchingRate.Vegetative.PotentialBranchingRate')
potential
```


## Modify a found model 
```r
new_model <- potential$node
new_model$XProperty <- 'NewVariable'
```

## Replace the new model 
```r
new <- replace_model(m, potential$path, new_model)
```

## Save into a new apsimx file
```r
write_apsimx(new, tempfile(fileext = '.json'))
```


## Run apsimx file

A function `run_models` is wrapped for APSIM NG `Models.exe` in the command line and can be called to run apsimx files. See [APSIM website](https://apsimnextgeneration.netlify.app/) for documentation. 

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----setup, message=FALSE, warning=FALSE--------------------------------------
library(rapsimng)
library(tidyverse)

## ----define-parameter---------------------------------------------------------
phyllochron_para <- tibble(parameter = "[Phenology].Phyllochron.BasePhyllochron.FixedValue", 
                           value = seq(60, 130, by = 1)) %>% 
  mutate(name = paste0("Cultivar", seq_len(n())))
head(phyllochron_para)

## -----------------------------------------------------------------------------
template <- read_apsimx(system.file("extdata/wheat.apsimx", package = "rapsimng"))

## -----------------------------------------------------------------------------
template <- update_cultivar(template, phyllochron_para)

node <- search_path(template, "[Permutation].Cv")    
if (length(node) == 0) {
  stop("[Permutation].Cv is not found")
}
new_value <- paste0("\\1", paste(phyllochron_para$name, collapse = ","))
node$node$Specification <- gsub("(.+ *= *)(.+)$", new_value, node$node$Specification)

node$node$Specification

template <- replace_model(template, node$path, node$node)

## -----------------------------------------------------------------------------

# write_apsimx(template, "new-path.apsimx")
# models_path <- "path-to-Models.exe"
# run_models(models_path, sim_name, csv = TRUE, parallel = FALSE)

## ----include = FALSE, warning=FALSE, error=FALSE, message=FALSE---------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----setup, echo=FALSE, warning=FALSE, error=FALSE, message=FALSE-------------
library(rapsimng)
suppressPackageStartupMessages(library(magrittr))

## ----read-apsimx--------------------------------------------------------------
library(rapsimng)
wheat <- read_apsimx(system.file("extdata/Wheat.json", package = "rapsimng"))
#wheat <- read_apsimx("inst/Wheat.json")

## -----------------------------------------------------------------------------
# Find the ThermalTime model 
a <- search_path(wheat, '[Wheat].Phenology.ThermalTime')
a$node$Children[[1]]$X[[2]]
# Update the optimum temperature
a$node$Children[[1]]$X[[2]] <- 27
# Replace with new value
wheat_new <- replace_model(wheat, a$path, a$node)
b <- search_path(wheat_new, '[Wheat].Phenology.ThermalTime')
# The optimum temperature should be updated now
b$node$Children[[1]]$X[[2]]

# The ThermalTime model can also be removed
a <- search_path(wheat, '[Wheat].Phenology.ThermalTime')
wheat_new <- remove_model(wheat, a$path)
b <- search_path(wheat_new, '[Wheat].Phenology.ThermalTime')
# The ThermalTime model should not be found now (i.e. Empty list)
b

## -----------------------------------------------------------------------------
new_cultivar <- new_model("PMF.Cultivar", name = "Hartog")
new_cultivar

## -----------------------------------------------------------------------------
new_cultivar$Command <- list(
  "[Phenology].MinimumLeafNumber.FixedValue = 6",
  "[Phenology].VrnSensitivity.FixedValue = 0")

## -----------------------------------------------------------------------------
# Read the apsimx file
wheat <- read_apsimx(system.file("extdata/wheat.apsimx", package = "rapsimng"))
# Create a new Replacements
replacements <- new_model("Core.Replacements")
# Insert the replacements into root folder
wheat_new <- insert_model(wheat, 1, replacements)
replacements_node <- search_path(wheat_new, ".Simulations.Replacements")
replacements_node$path

# Insert the new cultivar
wheat_new <- insert_model(wheat_new, replacements_node$path, new_cultivar)

# Check the new cultivar
cultivar_node <- search_path(wheat_new,
                                        ".Simulations.Replacements.Hartog")
cultivar_node$path
cultivar_node$node$Command
    

## -----------------------------------------------------------------------------
wheat <- read_apsimx(system.file("extdata/wheat.apsimx", package = "rapsimng"))
# Update cultivars
df <- data.frame(name = rep("Hartog", 3),
                  parameter = c("[Phenology].MinimumLeafNumber.FixedValue",
                               "[Phenology].VrnSensitivity.FixedValue",
                               "[Phenology].PpSensitivity.FixedValue"),
                 value = c(9, 7, 3))
wheat_cultivar <- update_cultivar(wheat, df)
# Check update cultivar paramters
hartog <- search_path(wheat_cultivar, "[Replacements].Hartog")
hartog$path

#' Test whether all files under published folder of apsimx are required
#'
#' @param base the base folder path to apsimx publish
#' @param example an example apsimx file
#'
#' @return A vector a required files
test_apsimx <- function(base, example) {
    # Sys.setenv(PATH = "")
    Sys.unsetenv("PATH")
    models_exe <- file.path(base, "Models.exe")
    all_files <- list.files(base, "*.dll$", full.names = TRUE, include.dirs = FALSE)

    required_files <- c("Models.exe")
    env <- new.env()
    assign("required_files", required_files, envir = env)

    all_files <- all_files[!(basename(all_files) %in% required_files)]

    new_files <- paste0(all_files, ".bak")
    i <- 1
    for (i in seq(along = all_files)) {
        file.rename(all_files[i], new_files[i])
        tryCatch(
            {
                rapsimng::run_models(models_exe, example)
            },
            error = function(e) {
                message(all_files[i], " is required.")
                required_files <- get("required_files", envir = env)
                required_files2 <- c(required_files, all_files[i])
                assign("required_files", required_files2, envir = env)
            }
        )
        file.rename(new_files[i], all_files[i])
    }

    required_files <- get("required_files", envir = env)
    unrequired_files <- all_files[!(all_files %in% required_files)]
    message("Unrequired files: \"", paste0(unrequired_files, collapse = ", "), "\"")
    required_files
}


#' Install ApsimNG Software
#'
#' This function installs the ApsimNG simulation software on the local system.
#' It handles the download and installation process for the APSIM Next Generation
#' agricultural modeling platform.
#'
#' @param repo Character string specifying the repository of ApsimNG to install.
#' @param branch Character string specifying the branch of the repository to install.
#' @param install_dir Character string specifying the directory where ApsimNG
#'   should be installed.
#' @param overwrite Logical. If TRUE, overwrites any existing installation
#'   at the specified install directory.
#' @return No return value.
#'
#' @details
#' This function downloads and installs ApsimNG from the official repository.
#' It performs version checking, handles dependencies, and configures the
#' installation for use with R. The function requires administrative privileges
#' on Windows systems.
#'
#' @examples
#' \dontrun{
#' # Install latest version
#' install_apsimng()
#' install_apsimng(install_dir = "C:/MyPrograms/ApsimNG")
#' }
#'
#' @export
install_apsimng <- function(
    repo = "https://github.com/APSIMInitiative/ApsimX.git",
    branch = "master",
    install_dir = "apsimx_build", 
    overwrite = FALSE
) {
    # Check arguments
    stopifnot(is.character(repo), length(repo) == 1)
    stopifnot(is.character(branch), length(branch) == 1)
    stopifnot(is.character(install_dir), length(install_dir) == 1)
    stopifnot(is.logical(overwrite), length(overwrite) == 1)

    # Check dependencies
    if (Sys.which("git") == "") {
        stop("Git is not available in PATH. Please install Git and ensure it's accessible.")
    }
    if (Sys.which("dotnet") == "") {
        stop(".NET SDK is not available in PATH. Please install .NET SDK and ensure it's accessible.")
    }

    # Set up directories
    target_dir <- file.path(tempdir(), paste0("apsim_ng_", Sys.getpid()))
    target_dir <- normalizePath(target_dir)
    if (dir.exists(target_dir)) {
        unlink(target_dir, recursive = TRUE, force = TRUE)
    }
    
    dir.create(target_dir, recursive = TRUE, showWarnings = FALSE)
    
    install_dir <- normalizePath(path.expand(install_dir), mustWork = FALSE)

    if (dir.exists(install_dir)) {
        warning("Install directory already exists: ", install_dir)
        if (!overwrite) {
            stop("Set 'overwrite = TRUE' to overwrite existing installation.")
        }
        unlink(install_dir, recursive = TRUE)
    }
    dir.create(install_dir, recursive = TRUE, showWarnings = FALSE)


    # Clone repo
    # git clone --filter=blob:none --sparse
    system2("git", c(
        "clone", "--depth=1", "--filter=blob:none", "--sparse",
        "-b", branch, repo, target_dir
    ))

    system2("git", c("-C", target_dir, "sparse-checkout", "set",
        "Models", "APSIM.Shared", "APSIM.Core", "DeepCloner.Core"
    ))


    message("Fetching APSIM NG source...")
    system2("git", c("clone", "--depth", "1", "-b", branch, repo, target_dir))

    # Build using .NET
    message("Compiling APSIM NG with .NET SDK...")
    csproj_path <- file.path(target_dir, "Models", "Models.csproj")
    result <- system2(
        "dotnet",
        c("build", csproj_path, "-c", "Release", "-o", install_dir),
        stdout = TRUE, stderr = TRUE
    )

    if (result != 0) stop("Build failed. Check .NET SDK and repo status.")

    # Clean build artifacts
    message("Cleaning temporary files...")
    unlink(target_dir, recursive = TRUE, force = TRUE)

    message("APSIM NG built successfully at: ", install_dir)
    return(invisible(install_dir))
}

#' Create a new model
#'
#' @param model The name of new model
#' @param name The new name
#'
#' @examples
#' new_model(model = "PMF.Cultivar")
#' new_model(model = "PMF.Cultivar", name = "example")
#' @export
new_model <- function(model, name = model) {
    if (length(model) != 1 || length(name) != 1) {
        stop("Only support 1 character value model")
    }
    # remove name space
    if (name == model) {
        name <- gsub("^([a-zA-Z_]+)(\\.*)(.*)$", "\\3", model)
        if (nchar(name) == 0) {
        name <- model
        }
    }
    models <- .get_models()
    # Find model
    m <- models[[model]]
    if (is.null(m)) {
        stop("Cannot find the models with name ", model)
    }
    m$Name <- name
    m
}


.get_models <- function() {
    define <- list(
        "PMF.Cultivar" = list(
            "$type" = "Models.PMF.Cultivar, Models",
            "Command" =  list(),
            "Name" = "Cultivar",
            "ResourceName" = NULL,
            "Children" = list(),
            "Enabled" = TRUE,
            "ReadOnly" = TRUE
        ), 
        "Core.Folder" = list(
            "$type" = "Models.Core.Folder, Models",
            "ShowInDocs" = FALSE,
            "Name" = "Folder",
            "ResourceName" = NULL,
            "Children" = list(),
            "Enabled" = TRUE,
            "ReadOnly" = FALSE
        )
    )
    define[["Core.Replacements"]] <- define[["Core.Folder"]]
    define[["Core.Replacements"]]$Name <- "Replacements"
    define[["PMF.CultivarFolder"]] <- define[["Core.Folder"]]
    define[["PMF.CultivarFolder"]]$Name <- "CultivarFolder"
    define
}

get_levels <- function(f) {
  r <- NULL

  if (!is.null(f$Specification) && f$Specification != "" &&
      !grepl('^\\[.+\\]$', f$Specification)) {
    pattern <- "^(\\[.+\\]\\..+ *= *)(.+)$"
    if (!grepl(pattern, f$Specification)) {
      stop("Specification should be as [SowingRule].Script.SowingDate = 2003-12-22")
    }
    levels <- gsub(pattern, "\\2", f$Specification)
    if (nchar(levels) == 0) {
      stop("No levels specified")
    }
    r <- strsplit(levels, ", *")[[1]]
  } else if (length(f$Children) > 0) {
    r <- unlist(lapply(f$Children, function(x) x$Name))
  } else {
    stop("Not supported factor")
  }
  r <- gsub("\\n", "", r)
  r <- gsub("\\r", "", r)
  return(r)
}

keep_level <- function(f, v) {
  # Check the value
  levels <- get_levels(f)
  pos <- !(v %in% levels)
  if (sum(pos) > 0) {
    stop("The value(s) (", paste(v[pos], collapse = ","),
         ") is not in the levels (", paste(levels, collapse = ", "), ")")
  }
  # In case of specification for levels
  if (!is.null(f$Specification) && f$Specification != "" &&
      !grepl('^\\[.+\\]$', f$Specification)) {
    pattern <- "^(\\[.+\\]\\..+ *= *)(.+)$"
    if (!grepl(pattern, f$Specification)) {
      stop("Specification should be as [SowingRule].Script.SowingDate = 2003-12-22")
    }
    specification_new <- gsub(pattern, "\\1", f$Specification)
    specification_new <- paste0(specification_new, paste(v, collapse = ", "))
    f$Specification <- specification_new
    # In case of multiple children
  } else if (length(f$Children) > 0) {
    new_children <- list()
    for (i in seq(along = f$Children)) {
      if (f$Children[[i]]$Name %in% v) {
        new_children[[length(new_children) + 1]] <- f$Children[[i]]
      }
    }
    f$Children <- new_children
  } else {
    stop("Not supported factor")
  }
  return(f)
}

get_factors <- function(l) {
  if (is.null(l$`$type`)) {
    stop("No $type element in the list")
  }
  factors <- NULL
  path <- 1
  if (l$`$type` == "Models.Factorial.Permutation, Models") {
    factors <- l
  } else if (l$`$type` == "Models.Factorial.Factors, Models") {
    # Find the permutation in the children
    perm <- search_path(l, "[Permutation]")
    if (length(perm) > 0) {
      factors <- perm$node
      path <- perm$path
    } else {
      if (length(l$Children) > 1) {
        stop("Not support for multiple factors in non permutation test.")
      }
      factors <- l
    }
  } else {
    stop("Not supported models type: ", l$`$type`)
  }
  if (length(factors) < 1) {
    stop("Not factors")
  }
  list(factors = factors, path = path)
}

#' Get simulations for a factorial experiment
#'
#' @param l A list from read_apsim with Factorial.Permutation as root.
#'
#' @return A list with Factor as name and Levels as values
#' @export
#' @examples
#' wheat <- read_apsimx(system.file("extdata/wheat.apsimx", package = "rapsimng"))
#' permutation <- search_path(wheat, path = "[Factors].Permutation")
#' get_simulations(permutation$node)
get_simulations <- function(l) {
  factors <- get_factors(l)$factors$Children
  res <- list()
  for (i in seq(along = factors)) {
      if (factors[[i]]$`$type` == "Models.Memo, Models") {
        next
      }
      res[[as.character(factors[[i]]$Name)]] <- get_levels(factors[[i]])
  }
  res
}

#' Keep simulations for a factorial experiment
#'
#' @param l A list from read_apsim with Factorial.Permutation as root.
#' @param s a list with factor as name and levels as value to keep.
#' The factor is kept if it is not specified.
#' @return A new list with removed simulations.
#' @export
#' @examples
#' wheat <- read_apsimx(system.file("extdata/wheat.apsimx", package = "rapsimng"))
#' permutation <- search_path(wheat, path = "[Factors].Permutation")
#' permutation_new <- keep_simulations(permutation$node, list(V = "2"))
#' get_simulations(permutation_new)
#'
#' permutation_new <- keep_simulations(permutation$node, list(Cv = c("Axe", "Bolac")))
#' get_simulations(permutation_new)
#'
#' permutation_new <- keep_simulations(permutation$node,
#'                                     list(V = "1", Cv = c("Axe", "Bolac")))
#' get_simulations(permutation_new)
keep_simulations <- function(l, s) {
  if (!methods::is(s, "list")) {
    stop("s requires as a list")
  }

  s_names <- names(s)
  pos <- is.null(s_names)
  if (sum(pos) > 0) {
    stop("All items of s should be named.")
  }

  info <- get_factors(l)
  factors_node <- info$factors
  factors <- factors_node$Children
  path <- info$path
  s_names <- names(s)
  for (i in seq(along = factors)) {
    if (factors[[i]]$Name %in% s_names) {
      factors[[i]] <- keep_level(factors[[i]], s[[factors[[i]]$Name]])
    } else {
      #l$Children[[i]] <- NULL
    }
  }
  factors_node$Children <- factors
  replace_model(l, path, factors_node)
}


#' Get the met file name for an experiment
#'
#' @param l A list or apsimxNode red by read_apsimx
#' @param is_stop Whether stop the function when error
#' @return The met file name in a experiment
#' @export
#'
#' @examples
#' wheat <- read_apsimx(system.file("extdata/wheat.apsimx", package = "rapsimng"))
#' exp <- search_path(wheat, path = "[Experiment]")
#' get_metfile(exp)
get_metfile <- function(l, is_stop = TRUE) {
  l <- get_node(l)

  met_node <- search_path(l, "[Weather]")
  if (length(met_node) == 0) {
      if (is_stop) {
          stop("Weather node is not found.")
      } else {
          warning("Weather node is not found.")
          return(NULL)
      }
  }
  if (met_node$node$`$type` != "Models.Climate.Weather, Models") {
    stop("Weather node is not a Models.Climate.Weather")
  }
  met_name <- met_node$node$FileName
  if (is.null(met_name)) {
      if (is_stop) {
         stop("Cannot found element FileName")
      } else {
         warning("Cannot found element FileName")
         return(NULL)
      }
  }
  met_name
}

suppressPackageStartupMessages(library(testthat))
test_that("Search all nodes", {
    # Read wheat.apsimx
    wheat <- read_apsimx(system.file("extdata/wheat.apsimx", package = "rapsimng"))

    permutation <- search_path(wheat, path = "[Experiment].Factors")
    expect_equal(length(permutation), 2)
    # Get simulations
    sims <- get_simulations(permutation$node)
    expect_equal(length(sims), 3)
    expect_equal(sims$V, c("1", "2"))
    expect_equal(sims$Cv, c("Axe", "Bolac", "Calingiri"))
    expect_equal(sims$N, c("0", "200"))

    # Keep simulations
    # l <- permutation$node
    # s <- list(V = "1", Cv = c("Axe", "Bolac"),
    #      N = "200")
    #
    permutation_new <- keep_simulations(permutation$node, list(V = "2"))
    sims_new <- get_simulations(permutation_new)
    expect_equal(length(sims_new), 3)
    expect_equal(sims_new$V, c("2"))
    expect_equal(sims_new$Cv, c("Axe", "Bolac", "Calingiri"))
    expect_equal(sims_new$N, c("0", "200"))

    permutation_new <- keep_simulations(permutation$node, list(Cv = c("Axe", "Bolac")))
    sims_new <- get_simulations(permutation_new)
    expect_equal(length(sims_new), 3)
    expect_equal(sims_new$V, c("1", "2"))
    expect_equal(sims_new$Cv, c("Axe", "Bolac"))
    expect_equal(sims_new$N, c("0", "200"))

    permutation_new <- keep_simulations(permutation$node,
                                        list(V = "1", Cv = c("Axe", "Bolac")))
    sims_new <- get_simulations(permutation_new)
    expect_equal(length(sims_new), 3)
    expect_equal(sims_new$V, "1")
    expect_equal(sims_new$Cv, c("Axe", "Bolac"))
    expect_equal(sims_new$N, c("0", "200"))

    permutation_new <- keep_simulations(permutation$node,
                                        list(V = "1", Cv = c("Axe", "Bolac"),
                                             N = "200"))
    sims_new <- get_simulations(permutation_new)
    expect_equal(length(sims_new), 3)
    expect_equal(sims_new$V, "1")
    expect_equal(sims_new$Cv, c("Axe", "Bolac"))
    expect_equal(sims_new$N, c("200"))

    # Check errors
    expect_error(keep_simulations(permutation$node, list(Cv = c("Axe", "Janz"))))
    expect_error(keep_simulations(permutation, list(Cv = c("Axe", "Janz"))))
    expect_error(keep_simulations(list(1), list(Cv = c("Axe", "Janz"))))

    # Test met name
    exp <- search_path(wheat, path = "[Experiment]")
    metfile <- get_metfile(exp)
    expect_equal(metfile, "example.met")
})

#' Read APSIMX file
#'
#' @param path The file path or URL to apsimx file
#'
#' @return A list object of apsimx file
#' @export
#'
#' @examples
#' file <- system.file("extdata/wheat.apsimx", package = "rapsimng")
#' m <- read_apsimx(file)
read_apsimx <- function(path) {

    jsonlite::read_json(path, simplifyDataFrame = FALSE,
                        simplifyVector = FALSE, simplifyMatrix = FALSE)
}


#' Write APSIMX file
#'
#' @param l the list of apsimx file
#' @param file The file path to apsimx file
#'
#' @return A list object of apsimx file
#' @export
write_apsimx <- function(l, file) {
    jsonlite::write_json(l, file, pretty = TRUE, auto_unbox = TRUE, null = 'null')
}

test_that("get_pawc", {
    wheat <- read_apsimx(system.file("extdata/wheat.apsimx", package = "rapsimng"))
    
    # Test basic PAWC calculation with LL15
    pawc <- get_pawc(wheat)
    expect_true(is.numeric(pawc))
    expect_equal(pawc, 361.23)
    expect_equal(length(pawc), 1)
    
    # Test PAWC with wheat crop-specific LL
    pawc_wheat <- get_pawc(wheat, crop = "Wheat")
    expect_true(is.numeric(pawc_wheat))
    expect_equal(pawc_wheat, 246.015)
    
    # Test PAWC with depth limit
    pawc_1000 <- get_pawc(wheat, depth = 1000)
    expect_true(is.numeric(pawc_1000))
    expect_true(pawc_1000 > 0)
    expect_true(pawc_1000 < pawc)  # Should be less than total
    
    # Test PAWC with non-existent crop (should use LL15)
    expect_warning({
        pawc_other <- get_pawc(wheat, crop = "Maize")
    })
    
    # Test error handling - invalid soil path
    expect_error({
        get_pawc(wheat, soil_path = "[InvalidSoil]")
    })
    
    # Test error handling - invalid depth
    expect_error({
        get_pawc(wheat, depth = -100)
    })
    
    expect_error({
        get_pawc(wheat, depth = "invalid")
    })
})

suppressPackageStartupMessages(library(testthat))
test_that("log", {
    # Insert a model
    # Add a new replacements
    wheat <- read_apsimx(system.file("extdata/wheat.apsimx", package = "rapsimng"))
    expect_error(log_level(wheat, "Notfound"))

    expect_error(wheat2 <- log_level(wheat, c("Error", "All")))

    expect_no_error(wheat2 <-log_level(wheat, "Error"))


    summary_node <- search_node(wheat2, all = FALSE, "$type" = "Models.Summary, Models")
    expect_equal(summary_node$node$Verbosity, 0)

    expect_no_error(wheat2 <-log_level(wheat, "All"))
    summary_node <- search_node(wheat2, all = FALSE, "$type" = "Models.Summary, Models")
    expect_equal(summary_node$node$Verbosity, 100)

})

test_that("Get experiment by name", {
    wheat <- read_apsimx(system.file("extdata/wheat.apsimx", package = "rapsimng"))
    expect_silent({
        exp_node <- get_experiment(wheat, "Experiment")
    })
    expect_equal(length(exp_node), 2)
    expect_equal(class(exp_node), "apsimxNode")
    expect_equal(exp_node$node$Name, "Experiment")
})

test_that("Error when experiment not found", {
    expect_error(get_experiment("1", "Experiment"))
    
    wheat <- read_apsimx(system.file("extdata/wheat.apsimx", package = "rapsimng"))
    expect_error(get_experiment(wheat, "NonExistentExperiment"))
    expect_error(get_experiment(wheat, c(1)))
    expect_error(get_experiment(wheat, c("Exp1", "Exp2")))

})

.frost_heat_param_names <- c(
    "FrostLowTT", "FrostMaxReductionRatio", "FrostUpTT", "FrostMinReductionRatio",
    "FrostStartSensitiveGS", "FrostStartMostSensitiveGS",
    "FrostEndMostSensitiveGS", "FrostEndSensitiveGS",
    "HeatLowTT", "HeatMinReductionRatio", "HeatUpTT", "HeatMaxReductionRatio",
    "HeatStartSensitiveGS", "HeatStartMostSensitiveGS",
    "HeatEndMostSensitiveGS", "HeatEndSensitiveGS"
)



.resolve_frost_heat_params <- function(params, crop = NULL) {
    if (!is.list(params)) {
        stop("'params' must be a list.")
    }

    if (all(.frost_heat_param_names %in% names(params))) {
        return(params)
    }

    if (is.null(crop) || length(crop) != 1 || !is.character(crop)) {
        stop("'crop' must be provided when 'params' is a named list of crops.")
    }
    if (!crop %in% names(params)) {
        stop("Crop '", crop, "' is not available in 'params'.")
    }

    params[[crop]]
}


.validate_frost_heat_inputs <- function(temp_min, temp_max, growth_stage, params) {
    if (!is.numeric(temp_min) || !is.numeric(temp_max) || !is.numeric(growth_stage)) {
        stop("'temp_min', 'temp_max', and 'growth_stage' must be numeric vectors.")
    }
    if (!all(length(temp_min) == length(temp_max),
             length(temp_min) == length(growth_stage))) {
        stop("'temp_min', 'temp_max', and 'growth_stage' must have the same length.")
    }

    missing_params <- setdiff(.frost_heat_param_names, names(params))
    if (length(missing_params) > 0) {
        stop("Missing required parameter(s): ", paste(missing_params, collapse = ", "))
    }
}



.check_numeric_vector <- function(x, name) {
    if (!is.numeric(x)) {
        stop("'", name, "' must be a numeric vector.")
    }
    if (length(x) < 1) {
        stop("'", name, "' must not be empty.")
    }
    if (any(!is.finite(x))) {
        stop("'", name, "' contains non-finite values.")
    }
}


.check_equal_length <- function(...) {
    args <- list(...)
    arg_names <- names(args)
    lengths <- vapply(args, length, FUN.VALUE = integer(1))
    if (length(unique(lengths)) > 1) {
        stop("'", paste(arg_names, collapse = "', '"), "' must have the same length.")
    }
}


.check_single_crop_params <- function(params, required = .frost_heat_param_names) {
    if (!is.list(params)) {
        stop("'params' must be a list.")
    }
    missing_params <- setdiff(required, names(params))
    if (length(missing_params) > 0) {
        stop("Missing required parameter(s): ", paste(missing_params, collapse = ", "))
    }
    params
}


.stage_sensitivity <- function(
    growth_stage,
    start_sensitive,
    start_most_sensitive,
    end_most_sensitive,
    end_sensitive
) {
    tidyweather::interpolation_function(
        x = c(start_sensitive, start_most_sensitive, end_most_sensitive, end_sensitive),
        y = c(0, 1, 1, 0),
        values = growth_stage
    )
}


#' Frost and Heat Damage Functions
#'
#' Helper and wrapper functions to calculate frost and heat damage from
#' temperature and growth stage.
#' 
#' @details 
#' 
#' The damage functions are based on the APSIM NG Model [FrostHeatDamageFunctions](https://github.com/APSIMInitiative/ApsimX/blob/3787815684da28c13467942b9d886fa04f54d4c6/Models/Functions/FrostHeatDamageFunctions.cs). See full documentation [here](https://notes.apsimng.bangyou.me/docs/Models/Functions/FrostHeatDamageFunctions.html).
#'
#' @name frost_heat_damage
NULL




#' Default Parameters For Frost And Heat Damage
#'
#' @rdname frost_heat_damage
#'
#' @return A named list with crop-specific parameter lists.
#' @export
#' @examples
#' default_frost_heat_params()
default_frost_heat_params <- function() {
    list(
        Wheat = list(
            FrostLowTT = -4.0, FrostMaxReductionRatio = 0.4,
            FrostUpTT = 1.0, FrostMinReductionRatio = 0,
            FrostStartSensitiveGS = 6.4854544,
            FrostStartMostSensitiveGS = 8.0,
            FrostEndMostSensitiveGS = 9.4983302,
            FrostEndSensitiveGS = 9.4990961,
            HeatLowTT = 30.3459986, HeatMinReductionRatio = 0,
            HeatUpTT = 34.0000667, HeatMaxReductionRatio = 0.2031898,
            HeatStartSensitiveGS = 7.2832313,
            HeatStartMostSensitiveGS = 7.483232,
            HeatEndMostSensitiveGS = 8.9709024,
            HeatEndSensitiveGS = 9.0945434
        ),
        Canola = list(
            FrostLowTT = -4.108862, FrostMaxReductionRatio = 0.3050003,
            FrostUpTT = -0.7549879, FrostMinReductionRatio = 0,
            FrostStartSensitiveGS = 7.222864,
            FrostStartMostSensitiveGS = 7.329504,
            FrostEndMostSensitiveGS = 8.063452,
            FrostEndSensitiveGS = 8.181088,
            HeatLowTT = 27.45271, HeatMinReductionRatio = 0,
            HeatUpTT = 38.8354, HeatMaxReductionRatio = 0.5810167,
            HeatStartSensitiveGS = 5.578273,
            HeatStartMostSensitiveGS = 6.260291,
            HeatEndMostSensitiveGS = 9.200536,
            HeatEndSensitiveGS = 9.428058
        )
    )
}

#' Frost Stage Sensitivity
#'
#' @rdname frost_heat_damage
#'
#' @param growth_stage Numeric vector of daily growth stage.
#' @param params A crop parameter list.
#'
#' @return A numeric vector of frost stage sensitivity (0-1).
#' @export
stage_sens_frost <- function(growth_stage, params) {
    .check_numeric_vector(growth_stage, "growth_stage")
    params <- .check_single_crop_params(
        params,
        required = c(
            "FrostStartSensitiveGS", "FrostStartMostSensitiveGS",
            "FrostEndMostSensitiveGS", "FrostEndSensitiveGS"
        )
    )

    .stage_sensitivity(
        growth_stage,
        params$FrostStartSensitiveGS,
        params$FrostStartMostSensitiveGS,
        params$FrostEndMostSensitiveGS,
        params$FrostEndSensitiveGS
    )
}


#' Heat Stage Sensitivity
#'
#' @rdname frost_heat_damage
#'
#' @param growth_stage Numeric vector of daily growth stage.
#' @param params A crop parameter list.
#'
#' @return A numeric vector of heat stage sensitivity (0-1).
#' @export
stage_sens_heat <- function(growth_stage, params) {
    .check_numeric_vector(growth_stage, "growth_stage")
    params <- .check_single_crop_params(
        params,
        required = c(
            "HeatStartSensitiveGS", "HeatStartMostSensitiveGS",
            "HeatEndMostSensitiveGS", "HeatEndSensitiveGS"
        )
    )

    .stage_sensitivity(
        growth_stage,
        params$HeatStartSensitiveGS,
        params$HeatStartMostSensitiveGS,
        params$HeatEndMostSensitiveGS,
        params$HeatEndSensitiveGS
    )
}


#' Frost Temperature Damage Function
#'
#' @rdname frost_heat_damage
#'
#' @param mint Numeric vector of daily minimum temperature.
#' @param params A crop parameter list.
#'
#' @return A numeric vector of frost damage contribution from temperature.
#' @export
temp_damage_frost <- function(mint, params) {
    .check_numeric_vector(mint, "mint")
    params <- .check_single_crop_params(
        params,
        required = c("FrostLowTT", "FrostUpTT", "FrostMaxReductionRatio", "FrostMinReductionRatio")
    )

    tidyweather::interpolation_function(
        x = c(params$FrostLowTT, params$FrostUpTT),
        y = c(params$FrostMaxReductionRatio, params$FrostMinReductionRatio),
        values = mint
    )
}


#' Heat Temperature Damage Function
#'
#' @rdname frost_heat_damage
#'
#' @param maxt Numeric vector of daily maximum temperature.
#' @param params A crop parameter list.
#'
#' @return A numeric vector of heat damage contribution from temperature.
#' @export
temp_damage_heat <- function(maxt, params) {
    .check_numeric_vector(maxt, "maxt")
    params <- .check_single_crop_params(
        params,
        required = c("HeatLowTT", "HeatUpTT", "HeatMinReductionRatio", "HeatMaxReductionRatio")
    )

    tidyweather::interpolation_function(
        x = c(params$HeatLowTT, params$HeatUpTT),
        y = c(params$HeatMinReductionRatio, params$HeatMaxReductionRatio),
        values = maxt
    )
}


#' Daily Frost Damage
#'
#' @rdname frost_heat_damage
#'
#' @param mint Numeric vector of daily minimum temperature.
#' @param growth_stage Numeric vector of daily growth stage.
#' @param params A crop parameter list.
#'
#' @return A numeric vector of daily frost damage.
#' @export
daily_damage_frost <- function(mint, growth_stage, params) {
    .check_numeric_vector(mint, "mint")
    .check_numeric_vector(growth_stage, "growth_stage")
    .check_equal_length(mint = mint, growth_stage = growth_stage)
    params <- .check_single_crop_params(params)

    temp_damage_frost(mint, params) * stage_sens_frost(growth_stage, params)
}


#' Daily Heat Damage
#'
#' @rdname frost_heat_damage
#'
#' @param maxt Numeric vector of daily maximum temperature.
#' @param growth_stage Numeric vector of daily growth stage.
#' @param params A crop parameter list.
#'
#' @return A numeric vector of daily heat damage.
#' @export
daily_damage_heat <- function(maxt, growth_stage, params) {
    .check_numeric_vector(maxt, "maxt")
    .check_numeric_vector(growth_stage, "growth_stage")
    .check_equal_length(maxt = maxt, growth_stage = growth_stage)
    params <- .check_single_crop_params(params)

    temp_damage_heat(maxt, params) * stage_sens_heat(growth_stage, params)
}


#' Daily Combined Frost And Heat Damage
#'
#' @rdname frost_heat_damage
#'
#' @param mint Numeric vector of daily minimum temperature.
#' @param maxt Numeric vector of daily maximum temperature.
#' @param growth_stage Numeric vector of daily growth stage.
#' @param params A crop parameter list.
#'
#' @return A numeric vector of combined daily frost and heat damage.
#' @export
daily_damage_frost_heat <- function(mint, maxt, growth_stage, params) {
    .check_numeric_vector(mint, "mint")
    .check_numeric_vector(maxt, "maxt")
    .check_numeric_vector(growth_stage, "growth_stage")
    .check_equal_length(mint = mint, maxt = maxt, growth_stage = growth_stage)
    params <- .check_single_crop_params(params)

    frost_damage <- daily_damage_frost(mint, growth_stage, params)
    heat_damage <- daily_damage_heat(maxt, growth_stage, params)
    1 - (1 - frost_damage) * (1 - heat_damage)
}


#' Cumulative Frost Damage
#'
#' @rdname frost_heat_damage
#'
#' @param mint Numeric vector of daily minimum temperature.
#' @param growth_stage Numeric vector of daily growth stage.
#' @param params A crop parameter list.
#'
#' @return A scalar cumulative frost damage.
#' @export
cum_damage_frost <- function(mint, growth_stage, params) {
    .check_numeric_vector(mint, "mint")
    .check_numeric_vector(growth_stage, "growth_stage")
    .check_equal_length(mint = mint, growth_stage = growth_stage)
    params <- .check_single_crop_params(params)

    1 - prod(1 - daily_damage_frost(mint, growth_stage, params))
}


#' Cumulative Heat Damage
#'
#' @rdname frost_heat_damage
#'
#' @param maxt Numeric vector of daily maximum temperature.
#' @param growth_stage Numeric vector of daily growth stage.
#' @param params A crop parameter list.
#'
#' @return A scalar cumulative heat damage.
#' @export
cum_damage_heat <- function(maxt, growth_stage, params) {
    .check_numeric_vector(maxt, "maxt")
    .check_numeric_vector(growth_stage, "growth_stage")
    .check_equal_length(maxt = maxt, growth_stage = growth_stage)
    params <- .check_single_crop_params(params)

    1 - prod(1 - daily_damage_heat(maxt, growth_stage, params))
}


#' Cumulative Combined Frost And Heat Damage
#'
#' @rdname frost_heat_damage
#'
#' @param mint Numeric vector of daily minimum temperature.
#' @param maxt Numeric vector of daily maximum temperature.
#' @param growth_stage Numeric vector of daily growth stage.
#' @param params A crop parameter list.
#'
#' @return A scalar cumulative frost and heat damage.
#' @export
cum_damage_frost_heat <- function(mint, maxt, growth_stage, params) {
    .check_numeric_vector(mint, "mint")
    .check_numeric_vector(maxt, "maxt")
    .check_numeric_vector(growth_stage, "growth_stage")
    .check_equal_length(mint = mint, maxt = maxt, growth_stage = growth_stage)
    params <- .check_single_crop_params(params)

    1 - (1 - cum_damage_frost(mint, growth_stage, params)) *
        (1 - cum_damage_heat(maxt, growth_stage, params))
}


#' Calculate Daily Frost And Heat Diagnostics
#'
#' @rdname frost_heat_damage
#'
#' @param mint Numeric vector of daily minimum temperature.
#' @param maxt Numeric vector of daily maximum temperature.
#' @param growth_stage Numeric vector of daily growth stage.
#' @param crop Crop name to use from `default_frost_heat_params()`.
#' @param params A crop parameter list. If provided, it will override the `crop` argument.
#'
#' @return A tibble with daily stage sensitivity, temperature damage, and daily
#'   frost/heat/combined damage.
#' @export
#' @examples
#' calc_daily_damage_frost_heat(
#'   mint = c(-2, 0, 4),
#'   maxt = c(29, 33, 36),
#'   growth_stage = c(7.0, 8.0, 9.0),
#'   crop = "Wheat"
#' )
calc_daily_damage_frost_heat <- function(
    mint, 
    maxt, 
    growth_stage, 
    crop = c("Wheat", "Canola"),
    params = NULL
) {
    .check_numeric_vector(mint, "mint")
    .check_numeric_vector(maxt, "maxt")
    .check_numeric_vector(growth_stage, "growth_stage")
    .check_equal_length(mint = mint, maxt = maxt, growth_stage = growth_stage)
    if (is.null(params)) {
        crop <- match.arg(crop)
        crop_params <- default_frost_heat_params()[[crop]]
    } else {
        crop_params <- params
    }
    .validate_frost_heat_inputs(mint, maxt, growth_stage, crop_params)

    tibble::tibble(
        mint = mint,
        maxt = maxt,
        growth_stage = growth_stage,
        stage_sens_frost = stage_sens_frost(growth_stage, crop_params),
        stage_sens_heat = stage_sens_heat(growth_stage, crop_params),
        temp_damage_frost = temp_damage_frost(mint, crop_params),
        temp_damage_heat = temp_damage_heat(maxt, crop_params),
        daily_damage_frost = daily_damage_frost(mint, growth_stage, crop_params),
        daily_damage_heat = daily_damage_heat(maxt, growth_stage, crop_params),
        daily_damage_frost_heat = daily_damage_frost_heat(mint, maxt, growth_stage, crop_params)
    )
}

#' Calculate Cumulative Frost And Heat Diagnostics
#'
#' @rdname frost_heat_damage
#'
#' @param mint Numeric vector of daily minimum temperature.
#' @param maxt Numeric vector of daily maximum temperature.
#' @param growth_stage Numeric vector of daily growth stage.
#' @param crop Crop name to use from `default_frost_heat_params()`.
#' @param params A crop parameter list. If provided, it will override the `crop` argument.
#'
#' @return A tibble with cumulative frost damage, cumulative heat damage, and
#'   cumulative combined frost-heat damage.
#' @export
#' @examples
#' calc_cum_damage_frost_heat(
#'   mint = c(-2, 0, 4),
#'   maxt = c(29, 33, 36),
#'   growth_stage = c(7.0, 8.0, 9.0),
#'   crop = "Wheat"
#' )
calc_cum_damage_frost_heat <- function(
    mint, 
    maxt, 
    growth_stage, 
    crop = c("Wheat", "Canola"),
    params = NULL
) {
    .check_numeric_vector(mint, "mint")
    .check_numeric_vector(maxt, "maxt")
    .check_numeric_vector(growth_stage, "growth_stage")
    .check_equal_length(mint = mint, maxt = maxt, growth_stage = growth_stage)
    crop <- match.arg(crop)
    if (is.null(params)) {
        crop_params <- default_frost_heat_params()[[crop]]
    } else {
        crop_params <- params
    }
    .validate_frost_heat_inputs(mint, maxt, growth_stage, crop_params)

    cum_frost <- cum_damage_frost(mint, growth_stage, crop_params)
    cum_heat <- cum_damage_heat(maxt, growth_stage, crop_params)
    cum_frost_heat <- cum_damage_frost_heat(mint, maxt, growth_stage, crop_params)
    tibble::tibble(
        cum_frost = cum_frost,
        cum_heat = cum_heat,
        cum_frost_heat = cum_frost_heat
    )
}

suppressPackageStartupMessages(library(testthat))

test_that("default_frost_heat_params returns expected crop list", {
    p <- default_frost_heat_params()
    expect_true(is.list(p))
    expect_true(all(c("Wheat", "Canola") %in% names(p)))
    expect_true(all(c("FrostLowTT", "HeatLowTT") %in% names(p$Wheat)))
})


test_that("stage_sens_frost returns valid values and validates input", {
    p <- default_frost_heat_params()$Wheat
    out <- stage_sens_frost(c(6.0, 8.0, 9.8), p)
    expect_equal(length(out), 3)
    expect_equal(out, c(0, 1, 0))
    expect_true(all(out >= 0 & out <= 1))
    expect_error(stage_sens_frost(c(6, NA), p))
})


test_that("stage_sens_heat returns valid values and validates input", {
    p <- default_frost_heat_params()$Wheat
    out <- stage_sens_heat(c(6.0, 8.0, 9.8), p)
    expect_equal(length(out), 3)
    expect_equal(out, c(0, 1, 0))
    expect_true(all(out >= 0 & out <= 1))

    expect_error(stage_sens_heat(c(6, Inf), p))
})


test_that("temp_damage_frost returns valid values and validates input", {
    p <- default_frost_heat_params()$Wheat
    out <- temp_damage_frost(c(-10, -2, 4), p)
    expect_equal(length(out), 3)
    expect_equal(out, c(0.4, 0.24, 0))
    expect_true(all(out >= 0 & out <= 1))

    expect_error(temp_damage_frost(character(0), p))
})


test_that("temp_damage_heat returns valid values and validates input", {
    p <- default_frost_heat_params()$Wheat
    out <- temp_damage_heat(c(25, 33, 40), p)
    expect_equal(length(out), 3)
    expect_true(all(out >= 0 & out <= 1))
    expect_equal(round(out, 4), c(0, 0.1476, 0.2032))
    expect_error(temp_damage_heat(c(25, NaN), p))
})


test_that("daily_damage_frost returns valid values and validates input", {
    p <- default_frost_heat_params()$Wheat
    out <- daily_damage_frost(c(-4, -2, 1), c(7, 8, 9), p)
    expect_equal(length(out), 3)
    expect_true(all(out >= 0 & out <= 1))
    expect_equal(round(out, 4), c(0.1359, 0.2400, 0))
    expect_error(daily_damage_frost(c(-4, -2), c(7), p))
    expect_error(daily_damage_frost(c(-4, -2, 1), c(7, 8, NA), p))
})


test_that("daily_damage_heat returns valid values and validates input", {
    p <- default_frost_heat_params()$Wheat
    out <- daily_damage_heat(c(30, 35, 38), c(7, 8, 9), p)
    expect_equal(length(out), 3)
    expect_true(all(out >= 0 & out <= 1))
    expect_equal(round(out, 4), c(0, 0.2032, 0.1554))
    expect_error(daily_damage_heat(c(30), c(7, 8), p))
})


test_that("daily_damage_frost_heat combines daily damage correctly", {
    p <- default_frost_heat_params()$Wheat
    mint <- c(-4, -2, 1)
    maxt <- c(30, 35, 38)
    gs <- c(7, 8, 9)

    df <- daily_damage_frost(mint, gs, p)
    dh <- daily_damage_heat(maxt, gs, p)
    out <- daily_damage_frost_heat(mint, maxt, gs, p)

    expect_equal(out, 1 - (1 - df) * (1 - dh))
    expect_error(daily_damage_frost_heat(c(-4), c(30, 35), c(7, 8), p))
})


test_that("cum_damage_frost returns scalar and expected formula", {
    p <- default_frost_heat_params()$Wheat
    mint <- c(-4, -2, 1)
    gs <- c(7, 8, 9)
    df <- daily_damage_frost(mint, gs, p)

    out <- cum_damage_frost(mint, gs, p)
    expect_equal(length(out), 1)
    expect_equal(out, 1 - prod(1 - df))
})


test_that("cum_damage_heat returns scalar and expected formula", {
    p <- default_frost_heat_params()$Wheat
    maxt <- c(30, 35, 38)
    gs <- c(7, 8, 9)
    dh <- daily_damage_heat(maxt, gs, p)

    out <- cum_damage_heat(maxt, gs, p)
    expect_equal(length(out), 1)
    expect_equal(out, 1 - prod(1 - dh))
})


test_that("cum_damage_frost_heat returns scalar and expected formula", {
    p <- default_frost_heat_params()$Wheat
    mint <- c(-4, -2, 1)
    maxt <- c(30, 35, 38)
    gs <- c(7, 8, 9)

    cf <- cum_damage_frost(mint, gs, p)
    ch <- cum_damage_heat(maxt, gs, p)
    out <- cum_damage_frost_heat(mint, maxt, gs, p)

    expect_equal(length(out), 1)
    expect_equal(out, 1 - (1 - cf) * (1 - ch))
})


test_that("calc_daily_damage_frost_heat returns expected columns", {
    out <- calc_daily_damage_frost_heat(
        mint = c(-4, 0, 5),
        maxt = c(25, 33, 40),
        growth_stage = c(6.5, 8.0, 9.3),
        crop = "Wheat"
    )

    expect_equal(nrow(out), 3)
    expect_true(all(c(
        "mint", "maxt", "growth_stage", "stage_sens_frost", "stage_sens_heat",
        "temp_damage_frost", "temp_damage_heat", "daily_damage_frost",
        "daily_damage_heat", "daily_damage_frost_heat"
    ) %in% names(out)))
    expect_error(calc_daily_damage_frost_heat(c(-1), c(30), c(7), crop = "Rice"))
})



test_that("calc_daily_damage_frost_heat with params", {
    out <- calc_daily_damage_frost_heat(
        mint = c(-4, 0, 5),
        maxt = c(25, 33, 40),
        growth_stage = c(6.5, 8.0, 9.3),
        params = default_frost_heat_params()$Wheat
    )

    expect_equal(nrow(out), 3)
    expect_true(all(c(
        "mint", "maxt", "growth_stage", "stage_sens_frost", "stage_sens_heat",
        "temp_damage_frost", "temp_damage_heat", "daily_damage_frost",
        "daily_damage_heat", "daily_damage_frost_heat"
    ) %in% names(out)))
})



test_that("calc_cum_damage_frost_heat returns expected columns", {
    out <- calc_cum_damage_frost_heat(
        mint = c(-4, 0, 5),
        maxt = c(25, 33, 40),
        growth_stage = c(6.5, 8.0, 9.3),
        crop = "Canola"
    )

    expect_equal(nrow(out), 1)
    expect_true(all(c("cum_frost", "cum_heat", "cum_frost_heat") %in% names(out)))
})


test_that("calc_cum_damage_frost_heat returns with params", {
    out <- calc_cum_damage_frost_heat(
        mint = c(-4, 0, 5),
        maxt = c(25, 33, 40),
        growth_stage = c(6.5, 8.0, 9.3),
        params = default_frost_heat_params()$Wheat
    )
    expect_equal(nrow(out), 1)
})

#' Set the log level of apsimx file
#'
#' @param l the list of apsimx file
#' @param level log level with option Error, Warning, Information, Diagnostic, All
#'
#' @return a new apsimx file
#' @export
log_level <- function(l, level = c("Error", "Warning", "Information", "Diagnostic", "All")) {
    level <- match.arg(level)
    Verbosity <- if (level == "Error") {
        0
    } else if (level == "Warning") {
        1
    } else if (level == "Information") {
        2
    } else if (level == "Diagnostic") {
        3
    } else if (level == "All") {
        100
    }

    summary_nodes <- search_node(l, all = TRUE, "$type" = "Models.Summary, Models")

    i <- 1
    for (i in seq(along = summary_nodes)) {
        summary_nodes[[i]]$node$Verbosity <- Verbosity
        l <- replace_model(l, summary_nodes[[i]]$path, summary_nodes[[i]]$node)
    }
    l
}

#' Create the minimum requirements to run an APSIM Next Generation
#'
#' @param install_path The installed path of APSIM Next Generation
#' @param output The output folder
#' @examples
#' \dontrun{
#'  minimum_apsimng("C:/ProgramFiles/APSIMNG", "minimum_apsimng")
#' }
#' @export
minimum_apsimng <- function(install_path, output) {
    if (!file.exists(output)) {
        dir.create(output)
    } else if (length(list.files(output, include.dirs = TRUE))) {
        stop("The output folder is not empty.")
    }
    base_files <- c("MathNet.Numerics.dll",
            "Models.exe",
            #"msvcr120.dll",
            "Newtonsoft.Json.dll",
            "sqlite3.dll",
            "zlib1.dll",
            "APSIM.Shared.dll",
            "FirebirdSql.Data.FirebirdClient.dll",
            "libsqlite3.so")
    files <- file.path(install_path, base_files)
    pos <- !file.exists(files)
    if (sum(pos) > 0) {
        stop("Files are not existed for ", paste(files[pos], collapse = ", "))
    }
    # Copy the APSIM files
    file.copy(files, file.path(output, basename(files)))

    if (file.exists('C:/Windows/SysWOW64/msvcr120.dll')) {
        file.copy('C:/Windows/SysWOW64/msvcr120.dll',
                  file.path(output, 'msvcr120.dll'))
    } else {
        warning('Cannot find C:/Windows/SysWOW64/msvcr120.dll. You have to manually copy it to model folder')
    }
    return(invisible())
}

#' Get the parent node from a path
#'
#' @param l the list of apsimx file
#' @param path If numeric, the path returned by search_path or search_node. If character, the path supported by apsimx
#'
#' @return A new list for parent
#' @export
#'
#' @examples
#' wheat <- read_apsimx(system.file("extdata/Wheat.json", package = "rapsimng"))
#' a <- search_path(wheat, '[Structure].BranchingRate')
#' get_parent(wheat, a$path)
get_parent <- function(l, path) {
    path <- .check_path(l, path)
    if (length(path) == 1) {
        stop("Already in the root for ", path)
    }
    if (length(path) == 2) {
        return (list(node = l, path = 1))
    }
    path <- path[-1]
    path <- path[-length(path)]
    eq <- 'l'
    for (i in seq(along = path)) {
        eq <- c(eq, '[["Children"]]', paste0('[[', path[i], ']]'))
    }
    eq_str <- paste(eq, collapse = '')
    eq_str <- paste0("l <- ", eq_str)
    eval(parse(text=eq_str))
    res <- list(node = l, path = c(1, path))
    res
}

#' List all reports in the database
#'
#' @param file the file path to apsimx or db file
#'
#' @return a vector of all reports
#' @export
#'
#' @examples
#' \dontrun{
#' file <- system.file("extdata/wheat.apsimx", package = "rapsimng")
#' list_report(file)
#' }
list_report <- function(file) {
    if (length(file) != 1) {
        stop("only support single file")
    }
    if (tools::file_ext(file) == "apsimx") {
        file <- paste0(tools::file_path_sans_ext(file), ".db")
    }
    if (tools::file_ext(file) != "db") {
        stop("Require a db file")
    }
    if (!file.exists(file)) {
        stop("File does not exist: ", file)
    }
    con <- DBI::dbConnect(RSQLite::SQLite(), file)
    tbls <- RSQLite::dbListTables(con)
    DBI::dbDisconnect(con)
    system_tbls <- c("_Checkpoints",
                     "_Factors", "_InitialConditions",
                     "_Messages", "_Simulations")
    tbls[!(tbls %in% system_tbls)]
}

#' Read apsimx database in db file format
#'
#' @param file the file path to apsimx or db file
#' @param report the report name
#'
#' @return a data.frame for a report
#' @export
#'
#' @examples
#' \dontrun{
#' file <- system.file("extdata/wheat.apsimx", package = "rapsimng")
#' read_report(file, "HarvestReport")
#' }
read_report <- function(file, report) {
    if (length(file) != 1) {
        stop("only support single file")
    }
    if (length(table) != 1) {
        stop("Only support single table")
    }
    if (tools::file_ext(file) == "apsimx") {
        file <- paste0(tools::file_path_sans_ext(file), ".db")
    }
    if (tools::file_ext(file) != "db") {
        stop("Require a db file")
    }
    if (!file.exists(file)) {
        stop("File does not exist: ", file)
    }
    con <- DBI::dbConnect(RSQLite::SQLite(), file)
    tbls <- RSQLite::dbListTables(con)
    if (!(report %in% tbls)) {
        message("The tables include: ", paste(tbls, collapse = ", "))
        DBI::dbDisconnect(con)
        stop("Cannot find table ", report, " in the database.")
    }
    simulations <- DBI::dbReadTable(con, "_Simulations") %>% tibble::tibble()
    table_con <- DBI::dbReadTable(con, report) %>% tibble::tibble()
    common_cols <- c("SimulationID" = "ID")
    if (rlang::has_name(table_con, "FolderName")) {
        common_cols <- c(common_cols, "FolderName")
    }
    table_con <- table_con %>%
        dplyr::left_join(simulations, by = common_cols)
    DBI::dbDisconnect(con)
    table_con
}

#' Split an APSIMX file into separate simulations except cultivar
#'
#' This function takes an APSIMX file and splits it into separate simulations for all factors except cultivar factors.
#' Then saving the results to the specified output location.
#'
#' @param file Character. Path to the input APSIMX file.
#' @param output Character. Path to the directory where the split components will be saved.
#'
#' @return No return value. The function is called for its side effects.
#'
#' @examples
#' \dontrun{
#' split_apsimx("path/to/input.apsimx", "path/to/output/directory")
#' }
#'
#' @export
split_apsimx <- function(file, output) {
    if (!is.character(file) || length(file) != 1) {
        stop("'file' must be a character string of length 1 (the path to the APSIMX file).")
    }
    if (!is.character(output) || length(output) != 1 || is.na(output) || !nzchar(output)) {
        stop("'output' must be a non-empty character string of length 1 (the path to the output directory).")
    }
    # Check if the output folder exists, and create it if it does not
    if (file.exists(output) && !dir.exists(output)) {
        stop(sprintf("A file named '%s' exists and is not a directory.", output))
    }
    if (!dir.exists(output)) {
        dir.create(output, recursive = TRUE)
    }
    if (!dir.exists(output)) {
        stop(sprintf("The output directory '%s' does not exist and could not be created.", output))
    }

    # read apsimx file
    apsimx_model <- read_apsimx(file)

    # Remove unused models
    apsimx_model_temp <- apsimx_model
    remove_nodes <- c(
        "[Simulations].Sensibility",
        "[Simulations].Sensibility tests",
        "[Simulations].Validation",
        "[Simulations].TitlePage",
        "[Simulations].Introduction",
        "[DataStore].ExcelMultiInput",
        "[DataStore].DailyObsPred",
        "[DataStore].HarvestObsPred",
        "[Replacements].Canola"
    )
    for (j in seq(along = remove_nodes)) {
        node_remove <- search_path(apsimx_model_temp, remove_nodes[j])
        if (length(node_remove) > 0) {
            apsimx_model_temp <- remove_model(apsimx_model_temp, node_remove$path)
        }
    }

    # Get all simulations

    sims <- search_node(apsimx_model, all = TRUE, `$type` = "Models.Core.Simulation, Models")
    sims <- sims |> purrr::discard(function(x) x$node$Name %in% c("CO2TEBaseSimulation", "WaterByNFactorial"))
    purrr::map_chr(sims, function(x) x$node$Name)

    for (j in seq(along = sims)) {
        # Check whether the parent is experiment
        parent <- get_parent(apsimx_model, sims[[j]]$path)
        if (parent$node$`$type` == "Models.Factorial.Experiment, Models") {
            exp_node <- parent$node
            exp_name <- exp_node$Name
            # obs_exp <- obs %>%
            #   filter(ExperimentName == exp_name)
            #
            # Remove all figures
            graphs <- search_node(exp_node,
                all = TRUE,
                `$type` = "Models.Graph, Models"
            )
            for (k in rev(seq(along = graphs))) {
                exp_node <- remove_model(exp_node, graphs[[k]]$path)
            }
            # Remove unused reports
            remove_nodes <- c(
                "[DailyReport]",
                "[MaxLeafSizeReport]",
                "[SowingReport]"
            )
            for (k in seq(along = remove_nodes)) {
                node_remove <- search_path(exp_node, remove_nodes[k])
                if (length(node_remove) > 0) {
                    exp_node <- remove_model(exp_node, node_remove$path)
                }
            }
            # Create a new apsimx file
            apsimx_model_exp <- apsimx_model_temp
            apsimx_model_exp <- insert_model(apsimx_model_exp, c(1), exp_node)
            # Get the met files
            met <- NULL
            met_file <- NULL
            tryCatch(
                {
                    met_file <- get_metfile(apsimx_model_exp)
                    met_path <- file.path(dirname(file), met_file)
                    met <- readLines(met_path)
                },
                error = function(e) {
                    # stop("met file is not found")
                }
            )
            # Further split apsimx by all factors except Cv
            all_nodes <- apsimx_model_exp %>%
                search_path("[Factors]")

            # find out all combinations of all other nodes
            simulations <- get_simulations(all_nodes$node)

            # # Check whether Cv node exists
            # pos <- names(simulations) %in% c("Cv", "Cultivar")
            # if (sum(pos) == 0) {
            #   warning("Cultivar factor doesn't exist for experiment ", exp_name)
            #   #next()
            # }
            # # Remove cultivar factor
            # simulations <- simulations[!(names(simulations) %in% c("Cv", "Cultivar"))]
            simulations <- expand.grid(simulations, stringsAsFactors = FALSE)
            m <- 1

            # Generate apsimx for each factor
            for (m in seq_len(max(1, nrow(simulations)))) {
                if (nrow(simulations) == 0) {
                    model_factor_new <- all_nodes$node
                    base_name <- exp_name
                } else {
                    model_factor_new <- keep_simulations(all_nodes$node, as.list(simulations[m, , drop = FALSE]))
                    factor_name <- paste(paste0(names(simulations), "_", simulations[m, ]), collapse = "_")
                    base_name <- paste0(exp_name, "-", factor_name)

                    factor_name <- paste(paste0(names(simulations), simulations[m, ]), collapse = "")
                    base_name <- paste0(exp_name, factor_name)
                    # obs_exp_f <- obs_exp
                    # mm <- 1
                    # for (mm in seq(along = names(simulations))) {
                    #   obs_exp_f %>%
                    #     filter(FactorName == names(simulations)[mm],
                    #            FactorValue == simulations[[mm]][m])
                    # }
                }
                base_name <- gsub(" ", "", base_name)
                apsimx_model_new <- apsimx_model_exp
                apsimx_model_new <- replace_model(
                    apsimx_model_exp,
                    all_nodes$path,
                    model_factor_new
                )
                file_name <- file.path(dirname(file), paste0(base_name, ".apsimx"))
                # write_apsimx(apsimx_model_new, file_name)
                # Combine weather and met file
                # apsimx <- readLines(file_name)

                file_name <- file.path(output, paste0(base_name, ".Rds"))
                # Check whether Weather node
                permutation_test <- model_factor_new$Children[[1]]$Children
                k <- 4
                for (k in seq(along = permutation_test)) {
                    if (!is.null(permutation_test[[k]]$Specification) &&
                        length(permutation_test[[k]]$Specification) == 1 &&
                        permutation_test[[k]]$Specification == "[Weather]") {
                        if (permutation_test[[k]]$Children[[1]]$`$type` == "Models.Climate.Weather, Models") {
                            met_file <- permutation_test[[k]]$Children[[1]]$FileName
                            met_path <- file.path(dirname(file), met_file)
                            new_met <- readLines(met_path)
                        }
                    }
                }
                if (is.null(met)) {
                    treat_node <- search_path(apsimx_model_new, "[Permutation].Treat")
                    if (length(treat_node) > 0) {
                        if (length(treat_node$node$Children) > 1) {
                            stop()
                        }
                        if (length(treat_node$node$Children[[1]]$Children) > 0) {
                            met_file <- treat_node$node$Children[[1]]$Children[[1]]$FileName
                            met_path <- file.path(dirname(file), met_file)
                            new_met <- readLines(met_path)
                        } else {
                            new_met <- NULL
                        }
                    }
                } else {
                    new_met <- met
                }
                saveRDS(list(met = new_met, apsimx = apsimx_model_new, met_file = met_file), file = file_name)
            }
        } else {
            # In case of single simulation
            apsimx_model_new <- apsimx_model_temp
            apsimx_model_new <- insert_model(apsimx_model_new, c(1), sims[[j]]$node)

            # Get weather data
            met <- NULL
            met_file <- NULL
            tryCatch(
                {
                    met_file <- get_metfile(apsimx_model_new)
                    met_path <- file.path(dirname(file), met_file)
                    met <- readLines(met_path)
                },
                error = function(e) {
                    # stop("met file is not found")
                }
            )
            base_name <- sims[[j]]$node$Name
            file_name <- file.path(output, paste0(base_name, ".Rds"))
            saveRDS(list(met = met, apsimx = apsimx_model_new, met_file = met_file), file = file_name)
        }
    }
}

# Internal code to check path
.check_path <- function(l, path) {
    if (length(path) < 1) {
        stop("Path should not be empty")
    }

    if (sum(path < 1) > 0) {
        stop("All path should be positive values")
    }
    if (sum(as.integer(path) != path) > 0){
        stop("All path should be integar values.")
    }
    if (path[1] != 1) {
        stop("Path should start with 1")
    }

    if (all(is.numeric(path))) {
        return(path)
    } else if (length(path) == 1 && is.character(path)) {
        return(search_path(l, path)$path)
    }
    stop("not implemented")
}


#' Replace a model with new values
#'
#' @param l the list of apsimx file
#' @param path If numeric, the path returned by search_path or search_node. If character, the path supported by apsimx
#' @param model A new model
#' @return The modified list with new value
#' @export
#' @examples
#' wheat <- read_apsimx(system.file("extdata/Wheat.json", package = "rapsimng"))
#'
#' a <- search_path(wheat, '[Wheat].Phenology.ThermalTime')
#' a$node$Children[[1]]$X[[2]] <- 27
#' wheat_new <- replace_model(wheat, a$path, a$node)
#' b <- search_path(wheat_new, '[Wheat].Phenology.ThermalTime')
#' b$node$Children[[1]]$X
replace_model <- function(l, path, model) {
    path <- .check_path(l, path)
    path <- path[-1]
    eq <- 'l'
    for (i in seq(along = path)) {
        eq <- c(eq, '[["Children"]]', paste0('[[', path[i], ']]'))
    }
    eq <- c(eq, '<- model')
    eq_str <- paste(eq, collapse = '')
    eval(parse(text=eq_str))
    return(l)
}



#' Disable models in apsimx
#'
#' @param l the list of apsimx file
#' @param paths If numeric, the path returned by search_path or search_node. If character, the path supported by apsimx
#' @return The modified list with new value
#' @export
#' @examples
#' wheat <- read_apsimx(system.file("extdata/Wheat.json", package = "rapsimng"))
#' a <- disable_models(wheat, '[Wheat].Phenology.ThermalTime')
disable_models <- function(l, paths) {
    i <- 1
    for (i in seq(along = paths)) {
        node_i <- search_path(l, paths[i])
        node_i$node$Enabled <- FALSE
        l <- replace_model(l, node_i$path, node_i$node)
    }
    return(l)
}



#' Remove a model with new values
#'
#' @param l the list of apsimx file
#' @param path If numeric, the path returned by search_path or search_node. If character, the path supported by apsimx
#' @return The modified list with new value
#' @export
#' @examples
#' wheat <- read_apsimx(system.file("extdata/Wheat.json", package = "rapsimng"))
#' a <- search_path(wheat, '[Wheat].Phenology.ThermalTime')
#' wheat_new <- remove_model(wheat, a$path)
#' b <- search_path(wheat_new, '[Wheat].Phenology.ThermalTime')
#' b
remove_model <- function(l, path) {
    path <- .check_path(l, path)
    path <- path[-1]
    eq <- 'l'
    for (i in seq(along = path)) {
        eq <- c(eq, '[["Children"]]', paste0('[[', path[i], ']]'))
    }
    eq <- c(eq, '<- NULL')
    eq_str <- paste(eq, collapse = '')
    eval(parse(text=eq_str))
    return(l)
}


#' Insert a model into apsimx
#'
#' @param l the list of apsimx file
#' @param path If numeric, the path returned by search_path or search_node. If character, the path supported by apsimx
#' @param model A new model
#' @return The modified list with new value
#' @export
#' @examples
#' wheat <- read_apsimx(system.file("extdata/wheat.apsimx", package = "rapsimng"))
#' replacements <- new_model("Core.Replacements")
#' wheat_new <- insert_model(wheat, 1, replacements)
#' replacements_node <- search_path(wheat_new, ".Simulations.Replacements")
#' replacements_node$path
#' # Add a cultivar folder under replacements
#' cultivar_folder <- new_model("PMF.CultivarFolder", "Cultivars")
#' wheat_new <- insert_model(wheat_new, replacements_node$path, cultivar_folder)
#' cultivar_folder_node <- search_path(wheat_new,
#'                                     ".Simulations.Replacements.Cultivars")
#' cultivar_folder_node$path
#' # Add an new cultivar
#' cultivar <- new_model("PMF.Cultivar", "Hartog")
#' wheat_new <- insert_model(wheat_new, cultivar_folder_node$path, cultivar)
#' cultivar_node <- search_path(wheat_new,
#'                              ".Simulations.Replacements.Cultivars.Hartog")
#' cultivar_node$path
insert_model <- function(l, path, model) {
    path <- .check_path(l, path)
    path <- path[-1]
    eq <- 'l'
    for (i in seq(along = path)) {
        eq <- c(eq, '[["Children"]]', paste0('[[', path[i], ']]'))
    }
    eq <- c(eq, '[["Children"]]')
    eq_str <- paste(eq, collapse = '')
    eq_str <- paste0(eq_str, '[[length(', eq_str, ')+1]] <- model')

    eval(parse(text=eq_str))
    return(l)
}



#' Insert models into apsimx
#'
#' @param l the list of apsimx file
#' @param path If numeric, the path returned by search_path or search_node. If character, the path supported by apsimx
#' @param models New models
#' @return The modified list with new value
#' @export
#' @examples
#' wheat <- read_apsimx(system.file("extdata/wheat.apsimx", package = "rapsimng"))
#' replacements <- new_model("Core.Replacements")
#' wheat_new <- insert_model(wheat, 1, replacements)
#' replacements_node <- search_path(wheat_new, ".Simulations.Replacements")
#' replacements_node$path
#' # Add a cultivar folder under replacements
#' cultivar_folder <- new_model("PMF.CultivarFolder", "Cultivars")
#' wheat_new <- insert_model(wheat_new, replacements_node$path, cultivar_folder)
#' cultivar_folder_node <- search_path(wheat_new,
#'                                     ".Simulations.Replacements.Cultivars")
#' cultivar_folder_node$path
#' # Add an new cultivar
#' cultivar <- new_model("PMF.Cultivar", "Hartog")
#' wheat_new <- insert_model(wheat_new, cultivar_folder_node$path, cultivar)
#' cultivar_node <- search_path(wheat_new,
#'                              ".Simulations.Replacements.Cultivars.Hartog")
#' cultivar_node$path
insert_models <- function(l, path, models) {
    path <- .check_path(l, path)
    path <- path[-1]
    eq <- 'l'
    for (i in seq(along = path)) {
        eq <- c(eq, '[["Children"]]', paste0('[[', path[i], ']]'))
    }
    eq <- c(eq, '[["Children"]]')
    eq_str <- paste(eq, collapse = '')
    #eq_str <- paste0(eq_str, '[[length(', eq_str, ')+1]] <- models')
    eq_str <- paste0(eq_str, "<- c(", eq_str, ', models)')
    eval(parse(text=eq_str))
    return(l)
}


#' append a model into apsimx
#'
#' @param l the list of apsimx file
#' @param path If numeric, the path returned by search_path or search_node. If character, the path supported by apsimx
#' @param model A new model which should be a list of new models
#' @return The modified list with new value
#' @export
#' @examples
#' wheat <- read_apsimx(system.file("extdata/wheat.apsimx", package = "rapsimng"))
#' replacements <- new_model("Core.Replacements")
#' wheat_new <- insert_model(wheat, 1, replacements)
#' replacements_node <- search_path(wheat_new, ".Simulations.Replacements")
#' replacements_node$path
#' # Add a cultivar folder under replacements
#' cultivar_folder <- new_model("PMF.CultivarFolder", "Cultivars")
#' wheat_new <- insert_model(wheat_new, replacements_node$path, cultivar_folder)
#' cultivar_folder_node <- search_path(wheat_new,
#'                                     ".Simulations.Replacements.Cultivars")
#' cultivar_folder_node$path
#' # Add an new cultivar
#' cultivar <- new_model("PMF.Cultivar", "Hartog")
#' wheat_new <- insert_model(wheat_new, cultivar_folder_node$path, cultivar)
#' cultivar_node <- search_path(wheat_new,
#'                              ".Simulations.Replacements.Cultivars.Hartog")
#' cultivar_node$path
#' # Append another cultivar
#' cultivar2 <- new_model("PMF.Cultivar", "Axe")
#' wheat_new <- append_model(wheat_new, cultivar_node$path, list(cultivar2))
#' cultivar2_node <- search_path(wheat_new,
#'                               ".Simulations.Replacements.Cultivars.Axe")
#' cultivar2_node$path
append_model <- function(l, path, model) {
    path <- .check_path(l, path)
    path <- path[-1]
    if (length(path) == 1) {
        stop("Path should be more than 1")
    }
    if (length(model) > 1) {
        stop("The length of new model should be equal to 1")
    }
    eq <- 'l'
    path1 <- path[-length(path)]
    path2 <- path[length(path)]
    for (i in seq(along = path1)) {
        eq <- c(eq, '[["Children"]]', paste0('[[', path[i], ']]'))
    }
    eq <- c(eq, '[["Children"]]')
    eq_str <- paste(eq, collapse = '')
    eq_str <- paste0(eq_str, ' <- append(', eq_str, ', model, ', path2, ')')

    eval(parse(text=eq_str))
    return(l)
}


#' Set a parameter with a new value
#'
#' @param l the list of apsimx file
#' @param parameter the name of parameter with APSIM NG specification
#' @param value the new value
#'
#' @return A list with replaced value
#' @export
#'
#' @examples
#'  wheat <- read_apsimx(system.file("extdata/Wheat.json", package = "rapsimng"))
#'  new_wheat <- set_parameter_value(wheat,
#'   "[Structure].BranchingRate.PotentialBranchingRate.Reproductive.Zero.FixedValue",
#'   1)
#' new_wheat2 <- search_path(new_wheat,
#'     "[Structure].BranchingRate.PotentialBranchingRate.Reproductive.Zero")
#' new_wheat2$node$FixedValue
#'
#' new_wheat <- set_parameter_value(
#'     wheat,
#'     "[Structure].HeightModel.WaterStress.XYPairs.Y",
#'     "0.1,1.1")
#' new_wheat2 <- search_path(new_wheat,
#'     "[Structure].HeightModel.WaterStress.XYPairs")
#' new_wheat2$node$Y
set_parameter_value <- function(l, parameter, value) {
    if (length(parameter) != 1) {
        stop("Require parameter has length 1")
    }
    if (length(value) != 1) {
        stop("Require value has length 1")
    }
    if (!is.character(parameter)) {
        stop("Require parameter is character")
    }
    if (grepl("\\.FixedValue$", parameter)) {
        p_path <- gsub("\\.FixedValue$", "", parameter)
        p_node <- search_path(l, p_path)
        if (length(p_node) == 0) {
            stop('Parameter (', parameter, ') is not found')
        }
        p_node$node$FixedValue <- value
    } else if (grepl("\\.(X|Y)$", parameter)) {
        p_path <- gsub("\\.(X|Y)$", "", parameter)
        p_node <- search_path(l, p_path)
        if (length(p_node) == 0) {
            stop('Parameter (', parameter, ') is not found')
        }
        new_values <- strsplit(value, " *, *")[[1]]

        if (grepl("\\.X$", parameter)) {
            if (length(new_values) != length(p_node$node$X)) {
                stop("New value doesn't match the length for old values. ",
                        "Expect: ", paste(p_node$node$X, collapse = ","))
            }
            p_node$node$X <- as.list(new_values)
        } else if (grepl("\\.Y$", parameter)) {
            if (length(new_values) != length(p_node$node$Y)) {
                stop("New value doesn't match the length for old values. ",
                        "Expect: ", paste(p_node$node$Y, collapse = ","))
            }
            p_node$node$Y <- as.list(new_values)
        }
    } else  {
        # for node with special value
        p_path <- gsub("^(.+)\\.([a-zA-Z0-9]+)$", "\\1", parameter)
        p_name <- gsub("^(.+)\\.([a-zA-Z0-9]+)$", "\\2", parameter)
        if (nchar(p_name) == 0 || nchar(p_name) == nchar(parameter)) {
            stop("Parameter name is not found")
        }
        p_node <- search_path(l, p_path)
        if (length(p_node) == 0) {
            stop('Parameter (', parameter, ') is not found')
        }
        p_node$node[[p_name]] <- value
    }
    l <- replace_model(l, p_node$path, p_node$node)
    l
}

suppressPackageStartupMessages(library(testthat))
test_that("cultivar", {
    wheat <- read_apsimx(system.file("extdata/Wheat.json", package = "rapsimng"))




    # Set a parameter value
    new_wheat <- set_parameter_value(wheat,
                                     "[Structure].BranchingRate.PotentialBranchingRate.Reproductive.Zero.FixedValue",
                                     1)
    new_wheat2 <- search_path(new_wheat, "[Structure].BranchingRate.PotentialBranchingRate.Reproductive.Zero")
    expect_equal(length(new_wheat2), 2)
    expect_equal(new_wheat2$node$FixedValue, 1)

    new_wheat <- set_parameter_value(wheat,
                                     "[Structure].HeightModel.WaterStress.XYPairs.X", "0.4,1.1")
    new_wheat2 <- search_path(new_wheat, "[Structure].HeightModel.WaterStress.XYPairs")
    expect_equal(length(new_wheat2), 2)
    expect_equal(new_wheat2$node$X[[1]], "0.4")

    new_wheat <- set_parameter_value(wheat,
                                     "[Structure].HeightModel.WaterStress.XYPairs.X", "0.4, 1.1")
    new_wheat2 <- search_path(new_wheat, "[Structure].HeightModel.WaterStress.XYPairs")
    expect_equal(length(new_wheat2), 2)
    expect_equal(new_wheat2$node$X[[1]], "0.4")

    new_wheat <- set_parameter_value(wheat,
                                     "[Structure].HeightModel.WaterStress.XYPairs.X", "0.4 , 1.1")
    new_wheat2 <- search_path(new_wheat, "[Structure].HeightModel.WaterStress.XYPairs")
    expect_equal(length(new_wheat2), 2)
    expect_equal(new_wheat2$node$X[[1]], "0.4")

    new_wheat <- set_parameter_value(wheat,
                                     "[Structure].HeightModel.WaterStress.XYPairs.X", "0.4 ,1.1")
    new_wheat2 <- search_path(new_wheat, "[Structure].HeightModel.WaterStress.XYPairs")
    expect_equal(length(new_wheat2), 2)
    expect_equal(new_wheat2$node$X[[1]], "0.4")

    expect_error({
        new_wheat <- set_parameter_value(wheat,
                                     "[Structure].HeightModel.WaterStress.XYPairs.X", "0.4")
        new_wheat <- set_parameter_value(wheat,
                                         "[Structure].HeightModel.WaterStress.XYPairs.X", "0.4,1.1,2")

    })

    # Disable models
    new_wheat <- disable_models(wheat, c("[Structure].HeightModel", "[Structure].BranchingRate"))



    new_wheat <- set_parameter_value(
        wheat,
        "[Structure].HeightModel.WaterStress.XYPairs.Y", "0.1,1.1")
    new_wheat2 <- search_path(new_wheat, "[Structure].HeightModel.WaterStress.XYPairs")
    expect_equal(length(new_wheat2), 2)
    expect_equal(new_wheat2$node$Y[[2]], "1.1")

    new_wheat <- set_parameter_value(
        wheat,
        "[Structure].CohortInitialisationStage", "Germination1")
    new_wheat2 <- search_path(new_wheat, "[Structure]")
    expect_equal(length(new_wheat2), 2)
    expect_equal(new_wheat2$node$CohortInitialisationStage, "Germination1")

    expect_error({
        new_wheat <- set_parameter_value(
            wheat,
            "[Structure].HeightModel.WaterStress.XYPairs.X", c(1, 1))
    })
    expect_error({
        new_wheat <- set_parameter_value(
            wheat,
            "[Structure].HeightModel.WaterStress", c(1, 1))
    })
    expect_error({
        new_wheat <- set_parameter_value(
            wheat, 1, c(1, 1))
    })

    expect_error({
        new_wheat <- set_parameter_value(
            wheat, c("A", "B"), c(1, 1))
    })





    # Replace a model
    a <- search_path(wheat, '[Wheat].Phenology.ThermalTime')
    wheat_new <- replace_model(wheat, a$path, a$node)
    expect_equal(wheat, wheat_new)

    a <- search_path(wheat, '[Wheat].Phenology.ThermalTime')
    a$node$Children[[1]]$X[[2]] <- 27
    wheat_new <- replace_model(wheat, a$path, a$node)
    b <- search_path(wheat_new, '[Wheat].Phenology.ThermalTime')
    expect_equal(b$node$Children[[1]]$X[[2]], 27)

    # Remove model
    a <- search_path(wheat, '[Wheat].Phenology.ThermalTime')
    wheat_new <- remove_model(wheat, a$path)
    b <- search_path(wheat_new, '[Wheat].Phenology.ThermalTime')
    expect_equal(b, list())

    # New model
    cultivar <- new_model("PMF.Cultivar")
    expect_equal(cultivar$Name, "Cultivar")
    expect_equal(cultivar$`$type`, "Models.PMF.Cultivar, Models")
    cultivar <- new_model("PMF.Cultivar", "test")
    expect_equal(cultivar$Name, "test")
    expect_equal(cultivar$`$type`, "Models.PMF.Cultivar, Models")



    # Insert a model
    # Add a new replacements
    wheat <- read_apsimx(system.file("extdata/wheat.apsimx", package = "rapsimng"))
    replacements <- new_model("Core.Replacements")
    wheat_replacement <- insert_model(wheat, 1, replacements)
    replacements_node <- search_path(wheat_replacement, ".Simulations.Replacements")
    expect_equal(length(replacements_node), 2)
    expect_equal(replacements_node$path, c(1, 3))
    # Add a cultivar folder under replacements
    cultivar_folder <- new_model("PMF.CultivarFolder", "Cultivars")
    wheat_cultivar_folder <- insert_model(wheat_replacement, replacements_node$path, cultivar_folder)
    cultivar_folder_node <- search_path(wheat_cultivar_folder,
                                        ".Simulations.Replacements.Cultivars")
    expect_equal(length(cultivar_folder_node), 2)
    expect_equal(cultivar_folder_node$path, c(1, 3, 1))

    # Add an new cultivar
    cultivar <- new_model("PMF.Cultivar", "Hartog")
    wheat_new_cultivar <- insert_model(wheat_cultivar_folder, cultivar_folder_node$path, cultivar)
    cultivar_node <- search_path(wheat_new_cultivar,
                                        ".Simulations.Replacements.Cultivars.Hartog")
    expect_equal(length(cultivar_node), 2)
    expect_equal(cultivar_node$path, c(1, 3, 1, 1))

    # Insert models
    cultivars <- c(list(new_model("PMF.Cultivar", "Hartog")),
                  list(new_model("PMF.Cultivar", "Janz")))
    wheat_new_cultivar2 <- insert_models(wheat_cultivar_folder, cultivar_folder_node$path, cultivars)

    cultivar_node <- search_path(wheat_new_cultivar2,
                                 ".Simulations.Replacements.Cultivars.Hartog")
    expect_equal(length(cultivar_node), 2)
    expect_equal(cultivar_node$path, c(1, 3, 1, 1))
    cultivar_node <- search_path(wheat_new_cultivar2,
                                 ".Simulations.Replacements.Cultivars.Janz")
    expect_equal(length(cultivar_node), 2)
    expect_equal(cultivar_node$path, c(1, 3, 1, 2))

    # Append node
    cultivar2 <- new_model("PMF.Cultivar", "Axe")
    wheat_new <- append_model(wheat_new_cultivar, cultivar_node$path, list(cultivar2))
    cultivar2_node <- search_path(wheat_new,
                                 ".Simulations.Replacements.Cultivars.Axe")
    expect_equal(length(cultivar2_node), 2)
    expect_equal(cultivar2_node$path, c(1, 3, 1, 2))


    # Update cultivars
    df <- rbind(data.frame(name = rep("Hartog", 3),
                     parameter = c("[Phenology].MinimumLeafNumber.FixedValue",
                                   "[Phenology].VrnSensitivity.FixedValue",
                                   "[Phenology].PpSensitivity.FixedValue"),
                     value = c(9, 7, 3)),
                data.frame(name = rep("Janz", 3),
                           parameter = c("[Phenology].MinimumLeafNumber.FixedValue",
                                         "[Phenology].VrnSensitivity.FixedValue",
                                         "[Phenology].PpSensitivity.FixedValue"),
                           value = c(3, 3, 4)))
    # No replacements
    wheat_cultivar <- update_cultivar(wheat, df)
    hartog <- search_path(wheat_cultivar, "[Replacements].Cultivars.Hartog")
    expect_equal(length(hartog), 2)
    expect_equal(hartog$path, c(1, 3, 1, 1))
    janz <- search_path(wheat_cultivar, "[Replacements].Cultivars.Janz")
    expect_equal(length(janz), 2)
    expect_equal(janz$path, c(1, 3, 1, 2))

    # Replacements and Cultivars node exist
    wheat_cultivar3 <- update_cultivar(wheat_new, df)
    hartog3 <- search_path(wheat_cultivar3, "[Replacements].Cultivars.Hartog")
    expect_equal(length(hartog3), 2)
    expect_equal(hartog3$path, c(1, 3, 1, 1))
    # Update existing one
    wheat_cultivar4 <- update_cultivar(wheat_cultivar3, df)
    hartog4 <- search_path(wheat_cultivar4, "[Replacements].Cultivars.Hartog")
    expect_equal(length(hartog4), 2)
    expect_equal(hartog3, hartog4)


})

#' Get an experiment node by name
#'
#' @param apsimx A parsed APSIMX model from `read_apsimx()`.
#' @param experiment A single experiment name to locate.
#'
#' @return An `apsimxNode` object for the experiment.
#' @export
#'
#' @examples
#' wheat <- read_apsimx(system.file("extdata/wheat.apsimx", package = "rapsimng"))
#' exp <- search_path(wheat, path = "[Experiment]")
#' get_experiment(wheat, exp$node$Name)
get_experiment <- function(apsimx, experiment) {
    stopifnot(is.list(apsimx) && length(apsimx) > 0)
    stopifnot(is.character(experiment) && length(experiment) == 1)
    experiment_node <- search_node(apsimx, "$type" = "Models.Factorial.Experiment, Models",
                "Name" = experiment)
    if (length(experiment_node) == 0) {
        stop("Experiment not found in the APSIMX file.")
    }
    return(experiment_node)
}

#' Estimate Plant Available Water Capacity (PAWC) from an apsimx file
#' 
#' This function calculates the Plant Available Water Capacity (PAWC) from soil 
#' properties in an APSIM Next Generation (.apsimx) file. PAWC represents the 
#' amount of water that can be stored in the soil and is available for plant uptake.
#' 
#' @details
#' 
#' PAWC represents the soil's water storage capacity between field capacity (DUL) and 
#' the permanent wilting point (LL). It is a critical parameter for:
#' 
#' PAWC is calculated for each soil layer using the formula:
#' 
#' **When crop = NULL and depth = NULL:**
#' \deqn{PAWC_i = (DUL_i - LL15_i) \times Thickness_i}
#' 
#' **When crop is specified:**
#' \deqn{PAWC_i = (DUL_i - LL_i) \times Thickness_i \times XF_i}
#' 
#' **When depth is specified (with or without crop):**
#' \deqn{PAWC_i = (DUL_i - LL_i) \times Thickness_i \times [XF_i] \times Weight_i}
#' 
#' Where:
#' \itemize{
#'   \item \strong{DUL} (Drained Upper Limit): The volumetric water content (mm/mm) 
#'         after the soil has drained under gravity, typically 2-3 days after saturation
#'   \item \strong{LL} (Lower Limit): The volumetric water content (mm/mm) below which 
#'         plants cannot extract water. This can be either:
#'         \itemize{
#'           \item \strong{LL15}: Generic lower limit at 15 bar suction (used when crop = NULL)
#'           \item \strong{Crop-specific LL}: Lower limit specific to a crop type, 
#'                 accounting for different rooting characteristics
#'         }
#'   \item \strong{Thickness}: Layer thickness in millimeters
#'   \item \strong{XF}: Exploration factor (0-1), representing the fraction of a layer 
#'         explored by roots. Only applied when crop is specified; represents the 
#'         proportion of the layer volume that roots can explore
#'   \item \strong{Weight}: Depth weighting factor (0-1), only applied when depth is specified:
#'         \itemize{
#'           \item 1.0 for layers completely within the specified depth
#'           \item Proportional fraction for layers partially within the depth
#'           \item 0.0 for layers completely beyond the specified depth
#'         }
#' }
#' 
#' The total PAWC is the sum across all layers:
#' \deqn{PAWC_{total} = \sum_{i=1}^{n} PAWC_i}
#' 
#' @param l The list of apsimx file
#' @param soil_path The path to the soil model. If NULL (default), searches for the 
#'   first Soil model. Can be a character string specifying the soil name.
#' @param crop The crop name for crop-specific lower limit (LL). 
#'   If NULL (default), uses LL15 and does not apply XF. If provided, uses the 
#'   crop-specific LL and XF values from the SoilCrop parameters.
#' @param depth The depth (in mm) to which PAWC should be calculated. 
#'   If NULL (default), calculates for all layers. If specified, applies proportional 
#'   weighting to layers that are partially or completely beyond this depth.
#'   
#' @return A numeric value of PAWC in mm
#' 
#' @export
#' 
#' @examples
#' wheat <- read_apsimx(system.file("extdata/wheat.apsimx", package = "rapsimng"))
#' 
#' # Get total PAWC using LL15 (generic)
#' pawc <- get_pawc(wheat)
#' pawc
#' 
#' # Get PAWC for wheat crop (uses crop-specific LL and XF)
#' pawc_wheat <- get_pawc(wheat, crop = "Wheat")
#' pawc_wheat
#' 
#' # Get PAWC to 1200 mm depth
#' pawc_1200 <- get_pawc(wheat, depth = 1200)
#' pawc_1200
#' 
#' # Get wheat PAWC to 1500 mm depth
#' pawc_wheat_1500 <- get_pawc(wheat, crop = "Wheat", depth = 1500)
#' pawc_wheat_1500
get_pawc <- function(l, soil_path = NULL, crop = NULL, depth = NULL) {
    if (is.null(soil_path)) {
        soil_node <- search_node(l, `$type` = "Models.Soils.Soil, Models")
    } else {
        stopifnot(is.character(soil_path) && length(soil_path) == 1)
        soil_node <- search_path(l, soil_path)
    }
    # Find the soil node
    
    if (length(soil_node) == 0) {
        stop("Soil node not found at path: ", soil_path)
    }
    
    # Find the Physical node
    physical_node <- search_node(soil_node$node, 
                                `$type` = "Models.Soils.Physical, Models")
    if (length(physical_node) == 0) {
        stop("Physical node not found in soil")
    }
    
    # Get soil properties
    thickness <- unlist(physical_node$node$Thickness)
    dul <- unlist(physical_node$node$DUL)
    ll15 <- unlist(physical_node$node$LL15)
    
    if (length(thickness) == 0 || length(dul) == 0 || length(ll15) == 0) {
        stop("Required soil properties (Thickness, DUL, LL15) not found")
    }
    
    # Check if all vectors have the same length
    if (!(length(thickness) == length(dul) && length(dul) == length(ll15))) {
        stop("Soil property vectors have different lengths")
    }
    
    # Determine which LL to use
    ll <- ll15
    xf <- rep(1, length(ll))  # Default XF = 1 for LL15
    # If crop is specified, look for crop-specific LL
    if (!is.null(crop)) {
        stopifnot(is.character(crop) && length(crop) == 1)
        # Find SoilCrop node for the specified crop
        soil_crops <- search_node(physical_node$node, 
                                    `$type` = "Models.Soils.SoilCrop, Models",
                                    all = TRUE)
        
        crop_found <- FALSE
        for (i in seq_along(soil_crops)) {
            if (grepl(tolower(crop), tolower(soil_crops[[i]]$node$Name))) {
                crop_ll <- unlist(soil_crops[[i]]$node$LL)
                stopifnot(length(crop_ll) == length(thickness))
                xf <- unlist(soil_crops[[i]]$node$XF)
                stopifnot(length(xf) == length(crop_ll))
                ll <- crop_ll
                crop_found <- TRUE
                break
            }
        }
        
        if (!crop_found) {
            warning("Crop '", crop, "' not found in SoilCrop. Using LL15 instead.")
        }
    }
    
    # Calculate cumulative depth
    cum_depth <- cumsum(thickness)
    
    # Create weight vector based on depth parameter
    weight <- rep(1, length(thickness))
    
    if (!is.null(depth)) {
        if (length(depth) != 1 || !is.numeric(depth) || depth <= 0) {
            stop("depth must be a positive number")
        }
        
        for (i in seq_along(cum_depth)) {
            if (i == 1) {
                depth_from <- 0
            } else {
                depth_from <- cum_depth[i - 1]
            }
            depth_to <- cum_depth[i]
            
            if (depth_from >= depth) {
                # Layer is completely beyond the specified depth
                weight[i] <- 0
            } else if (depth_to > depth) {
                # Layer is partially within the specified depth
                weight[i] <- (depth - depth_from) / thickness[i]
            }
            # else: Layer is completely within depth, weight remains 1
        }
    }
    
    # Calculate PAWC for each layer (mm)
    pawc_layer <- (dul - ll) * thickness * xf * weight
    
    # Return total PAWC
    total_pawc <- sum(pawc_layer)
    
    return(total_pawc)
}

#' Run apsimx file using Models.exe
#'
#' @param models_exe  path to Models.exe
#' @param path The path to an .apsimx file. May include wildcard.
#' @param pattern Use to filter simulation names to run.
#' @param recurse  Recursively search subdirectories for files matching ApsimXFileSpec. FALSE in default.
#' @param csv  Export all reports to .csv files. FALSE in default.
#' @param parallel Use the multi-process job runner. If FALSE, use single threaded; if TRUE, use the multi-process job runner
#' @param ncpus  Set the number of processors to use. All processes in default
#' @param verbose Write messages to StdOut when a simulation starts/finishes. Only has an effect when running a directory of .apsimx files (*.apsimx).
#'
#' @export
run_models <- function(models_exe, path,
                       pattern = NULL,
                       recurse = FALSE,
                       csv = FALSE,
                       parallel = NULL,
                       ncpus = NULL,
                       verbose = FALSE) {
    if (!file.exists(models_exe)) {
        stop(models_exe, " does not exist")
    }
    cmd <- paste0('"', models_exe, '" "', path, '"')
    # if (!is.null(pattern) && is.character(pattern)) {
    #     cmd <- paste0(cmd, " ", "--simulation-names ", pattern)
    # }
    if (is.logical(csv) && csv) {
        cmd <- paste0(cmd, " --csv")
    }
    if (verbose) {
        cmd <- paste0(cmd, " --verbose")
    }
    if (is.logical(recurse) && recurse) {
        cmd <- paste0(cmd, " --recursive")
    }
    if (is.logical(parallel) && !parallel) {
        cmd <- paste0(cmd, " --single-threaded ")
    }

    # if (is.logical(parallel) && parallel) {
    #     cmd <- paste0(cmd, " /MultiProcess")
    #     if (!is.null(ncpus) && is.numeric(ncpus) && ncpus > 0) {
    #         cmd <- paste0(cmd, " /NumberOfProcessors:", ncpus)
    #     }
    # }
    r <- system(cmd, intern = FALSE)
    if (r != 0) {
        stop("Failed to run apsimx: ", path)
    }
}


#' Run APSIM Next Generation Simulations in a Temporary Folder
#'
#' The `with_apsimx` function automates the process of setting up and running APSIM Next Generation (NG) simulations
#' in a temporary or specified folder. This approach helps reduce I/O overhead, particularly in high-performance
#' computing (HPC) environments with network file systems (NFS). The frequent file I/O operations performed by APSIM NG
#' (e.g., writing to SQLite databases) can strain file systems, and redirecting simulations to a local folder mitigates this issue.
#' The function can also perform optional post-processing of the simulation results.
#'
#' @param models A character string specifying the path to the APSIM NG executable (`Models.exe` on Windows
#'               or `Models` on Linux).
#' @param file A character string specifying the path to the `.apsimx` simulation file.
#' @param mets A character vector specifying paths to meteorological data (`.met` files) used in the simulation.
#'             Currently, only `.met` files located in the same folder as the `.apsimx` file are supported (optional).
#' @param target A character string specifying the target directory where simulations will be run.
#'               Defaults to the R system temporary directory (`tempdir()`).
#' @param clean A character string specifying which files or directories to clean after the simulation. Options are:
#'              \itemize{
#'                \item `"none"`: No files are cleaned (default).
#'                \item `"simulations"`: Cleans only simulation-related files (e.g., `*.apsimx`, `*.db`, `*.db-shm`, `*.db-wal`, `*.csv`).
#'                \item `"mets"`: Cleans only meteorological files (e.g., `*.met`).
#'                \item `"all"`: Cleans all files, including APSIM NG executable files, simulations, and meteorological files.
#'              }
#' @param post_process An optional function for post-processing simulation results.
#'                     The function must accept a `folder` argument specifying the directory
#'                     containing the simulation results.
#' @param ... Additional arguments passed to both the `run_models` and `post_process` functions.
#'
#' @return If a `post_process` function is provided, its return value is returned. Otherwise, the function returns `NULL`.
#'
#' @examples
#' \dontrun{
#' # Run simulations without post-processing
#' with_apsimx(
#'     models = "path/to/apsimx",
#'     file = "path/to/input.apsimx",
#'     mets = c("path/to/met1.met", "path/to/met2.met"),
#'     clean = "all"
#' )
#'
#' # Run simulations with post-processing
#' post_process_function <- function(folder) {
#'     output_files <- list.files(folder, full.names = TRUE)
#'     message("Output files:", paste(output_files, collapse = "\n"))
#' }
#'
#' result <- with_apsimx(
#'     models = "path/to/apsimx",
#'     file = "path/to/input.apsimx",
#'     mets = c("path/to/met1.met", "path/to/met2.met"),
#'     post_process = post_process_function
#' )
#' }
#'
#' @seealso
#' \code{\link{run_models}}: Runs APSIM simulations.
#' \code{\link{do.call}}: Dynamically calls functions with named arguments.
#'
#' @export
with_apsimx <- function(models, file, mets = NULL, target = tempdir(),
                        clean = c("none", "simulations", "mets", "all"),
                        post_process = NULL,
                        ...) {

    stopifnot(length(models) == 1)
    stopifnot(length(file) == 1)
    stopifnot(is.character(mets))
    stopifnot(is.character(target))
    stopifnot(file.exists(file))

    match.arg(clean)
    if ("none" %in% clean) {
        clean <- "none"
    }
    stopifnot(dir.exists(target))
    stopifnot(file.exists(models))

    if (!is.null(mets)) {
        if (!sum(file.exists(mets)) == length(mets)) {
            stop("Mets file is not existed")
        }
    }

    model_name <- if(.Platform$OS.type == "unix") {
        "Models"
    } else {
        "Models.exe"
    }

    # Create folder structure in the target folder (i.e. temp folder)
    # /apsimx for root folder

    folder_root <- file.path(target, "apsimx")
    dir.create(folder_root, showWarnings = FALSE)
    # Clean folder on exist
    if ("all" %in% clean) {
        on.exit(unlink(folder_root, recursive = TRUE))
    }
    stopifnot(file.exists(file.path(folder_root)))

    # /apsimx/models for apsimng Models.exe files
    folder_models <- file.path(folder_root, "models")
    dir.create(folder_models, showWarnings = FALSE)
    stopifnot(file.exists(file.path(folder_models)))

    # /apsimx/simulations for simulations
    folder_simulations <- file.path(folder_root, "simulation")
    dir.create(folder_simulations, showWarnings = FALSE)
    stopifnot(file.exists(file.path(folder_simulations)))
    # Clean simulations folder
    if ("simulations" %in% clean) {
        on.exit({
            old_files <- list.files(folder_simulations,
                                    "\\.(apsimx|db|db-shm|db-wal|csv)$",
                                    full.names = TRUE)
            a <- file.remove(old_files[file.exists(old_files)])
        })
    }
    if ("mets" %in% clean) {
        on.exit({
            old_files <- list.files(folder_simulations,
                                    "\\.(met)$",
                                    full.names = TRUE)
            a <- file.remove(old_files[file.exists(old_files)])
        })
    }

    # Copy execution files
    models_file <- file.path(folder_models, model_name)
    if (!file.exists(models_file)) {
        message("Copy ", model_name, " into ", folder_models)
        apsimx_files <- list.files(dirname(models), full.names = TRUE)
        a <- file.copy(apsimx_files, folder_models)
    }

    stopifnot(file.exists(models_file))

    # Copy apsimx and mets files
    file_new <- file.path(folder_simulations, basename(file))
    if (file.exists(file_new)) {
        warning("File is existed in target and overwrite: ", file_new)
    }
    file.copy(file, file_new, overwrite = TRUE)

    file_new_mets <- file.path(folder_simulations, basename(mets))
    if (sum(file.exists(file_new_mets)) > 0) {
        warning("Files are existed in target and ignore: ", file_new_mets)
    } else {
    # TO DO: Get mets file from apsimx and deal with subfolders
        a <- lapply(mets, file.copy, to = folder_simulations, overwrite = TRUE)
    }

    do_call2 <- function(fun, ...) {
        args <- list(...)
        args_names <- names(args)
        if (sum(nchar(args_names) == 0) > 0) {
            stop("All arguments should be named")
        }
        args_fun <- formals(fun)
        pos <- args_names %in% names(args_fun)
        a <- do.call(fun, args = args[pos])
        return(a)
    }
    do_call2(run_models, models_exe = models_file, path = file_new, ...)
    # post processing
    if (!is.null(post_process)) {
        stopifnot(is.function(post_process))
        res <- do_call2(post_process, folder = folder_simulations, ...)
        return(res)
    }
    return(invisible())
}

suppressPackageStartupMessages(library(testthat))
test_that("with_apsimx", {

    # Get models
    models <- Sys.getenv("APSIMNG_MODELS")
    skip_if(nchar(models) == 0)
    skip_if_not(file.exists(models))

    # Get example file
    file <- system.file("extdata/wheat.apsimx", package = "rapsimng")
    skip_if_not(file.exists(file))

    mets <- system.file("extdata/example.met", package = "rapsimng")
    skip_if_not(file.exists(mets))

    expect_error({
        a <- with_apsimx("A", file, mets)
    })

    expect_error({
        a <- with_apsimx(models, "AAA", mets)
    })
    expect_error({
        a <- with_apsimx(models, file, "AAA")
    })

    expect_error({
        a <- with_apsimx(models, file, mets, csv = TRUE, clean = "no-type")
    })

    expect_no_error({
        a <- with_apsimx(models, file, mets, csv = TRUE, clean = "all")
    })

    expect_no_error({
        a <- with_apsimx(models, file, mets, csv = TRUE, clean = "simulations")
    })
    expect_warning({
        a <- with_apsimx(models, file, mets, csv = TRUE, clean = "mets")
    })


    post_process <- function(folder, para1 = 10) {
        output_files <- list.files(folder, full.names = TRUE)
        message(output_files)
        harvest_report <- output_files[grepl("HarvestReport", output_files)]
        if (length(harvest_report) == 0) {
            return(NULL)
        }
        read.csv(harvest_report)
    }

    expect_no_error({
        res <- with_apsimx(models, file, mets, csv = TRUE, clean = "all", post_process = post_process)
    })

    expect_equal(nrow(res), 12)


    post_process <- function(folder, para1 = 10) {
        return(para1)
    }
    expect_no_error({
        res <- with_apsimx(models, file, mets, csv = TRUE, clean = "all", post_process = post_process)
    })
    expect_equal(res, 10)

    expect_no_error({
        res <- with_apsimx(models, file, mets, csv = TRUE, clean = "all", post_process = post_process, para1 = 100)
    })
    expect_equal(res, 100)
})

#' Get all cultivar parameters in a model
#'
#' @param l The list of apsimx file
#' @param alias Whether to export alias
#'
#' @return A data frame for all cultivar parameters
#' @export
#'
#' @examples
#' wheat <- read_apsimx(system.file("extdata/Wheat.json", package = "rapsimng"))
#' get_cultivar(wheat)
#' get_cultivar(wheat, alias = FALSE)
#'
get_cultivar <- function(l, alias = TRUE) {
    culs <- search_node(l, `$type` = "Models.PMF.Cultivar, Models", all = TRUE)
    if (length(culs) == 0) {
        stop("No cultivars found")
    }
    res <- list()
    k <- 0
    for (i in seq(along = culs)) {
        cul_name <- culs[[i]]$node$Name
        cmds <- unlist(culs[[i]]$node$Command)
        k <- k + 1
        res[[k]] <- data.frame(name = cul_name, 
                standard_name = cul_name,
                cmd = cmds,
                stringsAsFactors = FALSE)
        # for alias
        if (alias == TRUE && length(culs[[i]]$node$Children) > 0 ){
            aliases <- search_node(culs[[i]]$node, `$type` = 'Models.Core.Alias, Models', all = TRUE)
            if (length(aliases) > 0) {
                for (j in seq(along = aliases)) {
                    k <- k + 1
                    res[[k]] <- data.frame(name = aliases[[j]]$node$Name,
                            standard_name = cul_name,                    
                            cmd = cmds, stringsAsFactors = FALSE)
                }
            }
        }
    }
    if (length(res) == 0) {
        stop("No cultivars found")
    }
    r <- do.call(rbind, res)
    r <- r[grepl("=", r$cmd),]
    p_v <- strsplit(r$cmd, split = " *= *")
    r$parameter <- unlist(lapply(p_v, function(x) x[1]))
    r$value <- unlist(lapply(p_v, function(x) x[2]))
    r$cmd <- NULL
    r
}

# wheat <- read_apsimx(system.file("extdata/wheat.apsimx", package = "rapsimng"))
# df <- data.frame(name = rep("Hartog", 3),
#                  parameter = c("[Phenology].MinimumLeafNumber.FixedValue",
#                                "[Phenology].VrnSensitivity.FixedValue",
#                                "[Phenology].PpSensitivity.FixedValue"),
#                  value = c(9, 7, 3))
# l <- wheat
#' Title Update the cultivar parameters
#'
#' @description
#' This function assumes the file is apsimx format.
#' A new Replacements node is added if it is not exist.
#' The existing cultivar parameters are updated. New cultivar
#' is created.
#'
#'
#' @param l The list of apsimx file
#' @param df A data frame for new parameters with three columns, i.e. name, parameter and value.
#' @param add Whether to add extra nodes (e.g. replacements, Cultivars folder
#'  and new cultivar)
#' @param use_folder use cultivar folder to add new cultivars
#' @param cultivar_folder folder name for cultivars
#'
#' @return The modified apsimx file
#' @export
#'
#' @examples
#' wheat <- read_apsimx(system.file("extdata/wheat.apsimx", package = "rapsimng"))
#' # Update cultivars
#' df <- data.frame(name = rep("Hartog", 3),
#'                  parameter = c("[Phenology].MinimumLeafNumber.FixedValue",
#'                               "[Phenology].VrnSensitivity.FixedValue",
#'                               "[Phenology].PpSensitivity.FixedValue"),
#'                 value = c(9, 7, 3))
#'
#' wheat_cultivar <- update_cultivar(wheat, df)
#' hartog <- search_path(wheat_cultivar, "[Replacements].Hartog")
#' hartog$path
update_cultivar <- function(l, df, add = TRUE, use_folder = TRUE,
                            cultivar_folder = "Cultivars") {

    stopifnot(is.character(cultivar_folder))
    stopifnot(length(cultivar_folder) == 1)

    if (is.null(df$name)) {
        stop("'name' column is required in the data frame 'df'.")
    }

    if (is.null(df$parameter)) {
        stop("'parameter' column is required in the data frame 'df'.")
    }

    if (is.null(df$value)) {
        stop("'value' column is required in the data frame 'df'.")
    }

    if (add == TRUE) {
        cultivar_node <- search_node(l,
                                     `$type` = "Models.Core.Folder, Models",
                                     Name = cultivar_folder,
                                     all = FALSE)
        # compatible for old version
        if (length(cultivar_node) == 0) {
            cultivar_node <- search_node(l,
                                         `$type` = "Models.PMF.CultivarFolder, Models",
                                         Name = cultivar_folder,
                                         all = FALSE)

        }
        #cultivar_node <- search_path(l, "[Cultivars]")
        if (length(cultivar_node) == 0) {
            # Check the replacement node (assume the apsimx file)
            replacements_node <- search_path(l, "[Replacements]")

            if (length(replacements_node) == 0) {
                replacements <- new_model("Core.Replacements")
                l <- insert_model(l, 1, replacements)
            }
            if (use_folder) {
                replacements_node <- search_path(l, "[Replacements]")

                cultivars_path <- c(sprintf("[Replacements].%s", cultivar_folder))
                for (i in seq(along = cultivars_path)) {
                    cultivars_node <- search_path(l, cultivars_path[i])
                    if (length(cultivars_node) > 0) {
                        break
                    }
                }
                if (length(cultivars_node) == 0) {
                    # Fix me with new version
                    cultivar_model <- new_model("Core.Folder", cultivar_folder)
                    l <- insert_model(l, replacements_node$path, cultivar_model)
                }
            }


            # define the root node:
            # Cultivars or Replacements
            # Check whether cultivar node exists
            root_node <- search_path(l, "[Replacements]")
            if (use_folder) {
                root_node <- search_path(l, sprintf("[Replacements].%s", cultivar_folder))
            }
            # root_node <- search_path(l, "[Cultivars]")
            if (length(root_node) == 0) {
                stop("Cannot find replacements")
            }
        } else {
            root_node <- cultivar_node
        }
    }
    cultivars_name <- unique(df$name)
    i <- 1
    for (i in seq(along = cultivars_name)) {
        df_cultivar <- df[df$name == cultivars_name[i],]
        commands <- as.list(paste0(df_cultivar$parameter, " = ", df_cultivar$value))
        # Search whether the cultivar existing
        cultivar_node <- search_path(l, paste0("[", cultivars_name[i], "]"))
        if (length(cultivar_node) != 0 &&
            cultivar_node$node$`$type` == "Models.PMF.Cultivar, Models") {
            # overwrite in case the cultivar existing
            cultivar_model <- cultivar_node$node
            commands_old <- cultivar_model$Command
            for (j in seq(along = df_cultivar$parameter)) {
                pos <- grepl(df_cultivar$parameter[j], commands_old, fixed = TRUE)
                # Overwrite if existing
                if (sum(pos) > 0) {
                    commands_old[pos] <- commands[j]
                } else {
                    commands_old <- c(commands_old, commands[j])
                }
            }
            cultivar_model$Command <- commands_old
            # Replace cultivar
            l <- replace_model(l, cultivar_node$path,
                               cultivar_model)
        } else {
            if (add == TRUE) {
                # Create a new one if not existing
                cultivar_model <- new_model("PMF.Cultivar", cultivars_name[i])
                cultivar_model$Command <- commands
                # Insert new model
                l <- insert_model(l, root_node$path,
                                  cultivar_model)
            } else {
                warning("Cultivar ", cultivars_name[i], " cannot be found.")
            }
        }

    }
    l
}


#' Generate new cultivars with parameter which can be used in Replacements
#'
#' @param df A data frame for new parameters with three columns, i.e. name, parameter and value.
#' @param cultivar_folder folder name for cultivars
#'
#' @returns An APSIMX list
#' @export
#'
#' @examples
#' df <- data.frame(name = c("C1", "C1", "C2", "C2"),
#'      parameter = c("[Phenology].CAMP.FLNparams.MinLN",
#'               "[Phenology].CAMP.FLNparams.VrnLN",
#'               "[Phenology].CAMP.FLNparams.MinLN",
#'               "[Phenology].CAMP.FLNparams.VrnLN"),
#'      value = c(5, 6, 7, 8))
#' a <- new_cultivar(df)
new_cultivar <- function(df,
                         cultivar_folder = "Cultivars") {

    stopifnot(is.character(cultivar_folder))
    stopifnot(length(cultivar_folder) == 1)

    if (is.null(df$name)) {
        stop("'name' column is required in the data frame 'df'.")
    }

    if (is.null(df$parameter)) {
        stop("'parameter' column is required in the data frame 'df'.")
    }

    if (is.null(df$value)) {
        stop("'value' column is required in the data frame 'df'.")
    }
    cultivar_node <- new_model("Core.Folder", cultivar_folder)

    cultivar_path <- search_path(cultivar_node, paste0("[", cultivar_folder, "]"))
    cultivar_path <- cultivar_path$path

    cultivars_name <- unique(df$name)
    i <- 1
    for (i in seq(along = cultivars_name)) {
        df_cultivar <- df[df$name == cultivars_name[i],]
        commands <- as.list(paste0(df_cultivar$parameter, " = ", df_cultivar$value))
        # Search whether the cultivar existing

        # Create a new one if not existing
        cultivar_model <- new_model("PMF.Cultivar", cultivars_name[i])
        cultivar_model$Command <- commands
        # Insert new model
        cultivar_node <- insert_model(cultivar_node, cultivar_path,
                          cultivar_model)
    }
    cultivar_node
}

suppressPackageStartupMessages(library(testthat))
test_that("cultivar", {
    wheat <- read_apsimx(system.file("extdata/Wheat.json", package = "rapsimng"))
    a <- get_cultivar(wheat)
    expect_equal(nrow(a), 30)
    expect_equal(length(a$name), 30)
    expect_equal(length(a$standard_name), 30)
    expect_equal(length(a$parameter), 30)
    expect_equal(length(a$value), 30)
    a <- get_cultivar(wheat, alias = FALSE)
    expect_equal(nrow(a), 25)
})



test_that("new cultivar", {

    df <- data.frame(name = c("C1", "C1", "C2", "C2"),
                     parameter = c("[Phenology].CAMP.FLNparams.MinLN",
                                   "[Phenology].CAMP.FLNparams.VrnLN",
                                   "[Phenology].CAMP.FLNparams.MinLN",
                                   "[Phenology].CAMP.FLNparams.VrnLN"),
                     value = c(5, 6, 7, 8))
    cultivar_folder <- "Cultivars"
    a <- new_cultivar(df, cultivar_folder)
    expect_equal(a$Name, cultivar_folder)
    expect_equal(length(a$Children), 2)
    expect_equal(a$Children[[1]]$Name, "C1")
    expect_equal(a$Children[[1]]$Command[[1]], "[Phenology].CAMP.FLNparams.MinLN = 5")
    expect_equal(a$Children[[2]]$Name, "C2")
})

# * Author:    Bangyou Zheng (Bangyou.Zheng@csiro.au)
# * Created:   01:09 PM Thursday, 21 February 2019
# * Copyright: MIT


#' Find element(s) in  apsimx file
#'
#' @param l The list of apsimx file
#' @param all Whether to find all elements
#' @param max_depth The maximum depth to search
#' @param case_insensitive Whether case sensitive
#' @param ... Other names arguments for property to match
#'
#' @return A list matching all criteria if all equals to TRUE,
#' A list with node and path if all equals to FALSE (default)
#' @export
#'
#' @examples
#' wheat <- read_apsimx(system.file("extdata/Wheat.json", package = "rapsimng"))
#' # Return empty list if not found
#' search_node(wheat, Name = "Simulations1")
#' # Find root level
#' a <- search_node(wheat, Name = "Simulations")
#' a$path
#' # Find sub-level
#' a <- search_node(wheat, Name = "Wheat")
#' a$path
#' a <- search_node(wheat, `$type` = "Models.PMF.Cultivar, Models")
#' a$path
#'
#' # Find multiple attributes
#' a <- search_node(wheat,
#'             Name = 'PotentialBranchingRate',
#'             `$type` = "Models.Functions.PhaseLookup, Models")
#' a$path
#' a$node$Name
#' # Find all cultivar nodes
#'  a <- search_node(wheat, `$type` = "Models.PMF.Cultivar, Models", all = TRUE)
#' length(a)

search_node <- function(l, all = FALSE, max_depth = 1000000,
                        case_insensitive = TRUE, ...) {
    conds <- list(...)
    if (case_insensitive) {
        conds <- lapply(conds, tolower)
    }
    ele_names <- names(conds)

    if (!(!is.null(ele_names) && all(nchar(ele_names) > 0))) {
        stop('All elements should be named.')
    }
    r_search_node <- function(l, conds, ele_names, max_depth, cdepth,
                              cpath = 1, all = FALSE, result = list(),
                              case_insensitive = case_insensitive){
        check <- TRUE
        for (i in seq(along = conds)) {
            if (case_insensitive) {
                check <- check && !is.null(l[[ele_names[i]]]) &&
                    tolower(l[[ele_names[i]]]) == as.character(conds[i])

            } else {
                check <- check && !is.null(l[[ele_names[i]]]) &&
                    l[[ele_names[i]]] == as.character(conds[i])
            }

        }
        if (check) {
            # assign('l', l, .GlobalEnv)
            # stop()
            r <- list()
            r$node <- l
            r$path <- cpath[-length(cpath)]
            r$path <- cpath
            class(r) <- 'apsimxNode'
            result[[length(result) + 1]] <- r
            if (!all) {
                return (result)
            }
        }
        if (cdepth > max_depth) {
            return (result)
        }
        if (!is.null(l$Children) && length(l$Children) > 0) {
            for (i in seq(along = l$Children)) {

                result <- r_search_node(l$Children[[i]], conds, ele_names, max_depth, cdepth + 1,
                                     cpath = c(cpath, i), result = result,
                                     case_insensitive = case_insensitive)
            }
        }
        return (result)
    }
    res <- r_search_node(l, conds, ele_names, max_depth, 1, all = all,
                         case_insensitive = case_insensitive)
    if (length(res) == 0) {
        return (res)
    }
    if (!all) {
        if (length(all) != 1) {
            stop("Require 1 element if all = TRUE")
        }
        res <- res[[1]]
    }
    return(res)
}

#' Find a model in the apsimx file using specified path
#'
#' @section Absolute Paths:
#' Absolute paths have a leading ‘.’ e.g.
#' \itemize{
#'   \item .Simulations.Test.Clock - absolute path - refers to the clock model
#' in the 'Test' simulation.
#' }
#'
#' @section Scoped Paths:
#' Scoped paths have a leading model type in square brackets. A model
#' of the specified name, in scope, is located before applying the rest of the path.
#' \itemize{
#'    \item \[Soil\].Water - scoped path - refers to the Water model that is a child of a model
#' that has the name 'Soil' that is in scope
#' }
#'
#' @param l the list of apsimx file
#' @param path The specified path (See details)
#' @param case_insensitive Whether case sensitive

#' @return The list for the specified path.
#' @export
#'
#' @examples
#' wheat <- read_apsimx(system.file("extdata/Wheat.json", package = "rapsimng"))
#' # Return empty list if not found
#' search_path(wheat, "[Simulations1]")
#' # Search root path
#' a <- search_path(wheat, '.Simulations')
#' a$path
#' a$node$Name
#' # Level one
#' a <- search_path(wheat, '.Simulations.Wheat1')
#' a$path
#' a$node$Name
#' # Level two
#' a <- search_path(wheat, '.Simulations.Wheat')
#' a$path
#' a$node$Name
#' # Level three
#' a <- search_path(wheat, '.Simulations.Wheat.BranchingRate')
#' a$path
#' a$node$Name
#' a <- search_path(wheat, '.Simulations.Wheat.Structure')
#' a$path
#' a$node$Name
#' # Level four
#' a <- search_path(wheat, '.Simulations.Wheat.Structure.BranchingRate')
#' a$path
#' a$node$Name
#' a <- search_path(wheat, '.Simulations.Wheat.Structure.BranchingRate1')
#' a$path
#' a$node$Name
#' # scoped
#' # Root path
#' a <- search_path(wheat, '[Simulations1]')
#' a <- search_path(wheat, '[Simulations]')
#' a$path
#' a$node$Name
#' # Level two
#' a <- search_path(wheat, '[Simulations].Wheat1')
#' a <- search_path(wheat, '[Simulations1].Wheat')
#' a$path
#' a$node$Name
#' a <- search_path(wheat, '[Whea]')
#' a <- search_path(wheat, '[Wheat]')
#' a$path
#' a$node$Name
#' # Level three
#' a <- search_path(wheat, '[Wheat].BranchingRate')
#' a <- search_path(wheat, '[Wheat].Structure')
#' a$path
#' a$node$Name
#' a <- search_path(wheat, '[Structure]')
#' a$path
#' a$node$Name
#' # Level four
#' a <- search_path(wheat, '[Structure].BranchingRate')
#' a$path
#' a$node$Name
#' a <- search_path(wheat, '[Structure].BranchingRate1')
#' a <- search_path(wheat, '[Structure1].BranchingRate')
search_path <- function(l, path, case_insensitive = TRUE) {
    if (length(path) != 1) {
        stop('Only one path is supported.')
    }
    path_type <- ifelse(grepl('^\\.', path), 'absolute',
                        ifelse(grepl('^\\[.*\\]\\.*', path), 'scoped', 'NotSupported'))
    if (path_type == 'NotSupported') {
        stop('The path is not supported.')
    }
    if (path_type == 'scoped') {
        name <- sub('^\\[(.*)\\].*', '\\1', path)
        path <- sub('^\\[(.*)\\](.*)', '\\2', path)
        l <- search_node(l, Name = name, case_insensitive = case_insensitive)
        if (is.null(l$node)) {
            return(list())
        }
    } else if (path_type == 'absolute') {
        l <- list(node = l, path = c(1))
    }

    if (nchar(path) == 0) {
        return(l)
    }
    search_path_str <- sub('^\\.', '', path)
    search_path_str <- strsplit(search_path_str, '\\.')[[1]]

    current_node <- l
    for (i in seq_along(search_path_str)) {
        old_path <- current_node$path
        current_node <- search_node(current_node$node, Name = search_path_str[i], max_depth = 1,
                                    case_insensitive = case_insensitive)
        if (is.null(current_node$node)) {
            return (list())
        }
        current_node$path <- c(old_path, current_node$path[-1])
        if (is.null(current_node$node)) {
            return (list())
        }
    }
    return(current_node)
}

suppressPackageStartupMessages(library(testthat))
test_that("Search all nodes", {
    # Read wheat.json
    wheat <- read_apsimx(system.file("extdata/Wheat.json", package = "rapsimng"))
    # Search first node
    # Empty if no matching
    a <- search_node(wheat, Name = "Simulations1")
    expect_equal(length(a), 0)
    expect_equal(a, list())

    a <- search_node(wheat, Name = "simulations", case_insensitive = FALSE)
    expect_equal(length(a), 0)
    expect_equal(a, list())

    a <- search_node(wheat, Name = "Simulations")
    expect_equal(length(a), 2)
    expect_equal(a$path, 1)

    a <- search_node(wheat, Name = "simulations")
    expect_equal(length(a), 2)
    expect_equal(a$path, 1)


    a <- search_node(wheat, Name = "Wheat")
    expect_equal(length(a), 2)
    expect_equal(a$path, c(1, 1))

    a <- search_node(wheat, Name = "Arbitrator")
    expect_equal(length(a), 2)
    expect_equal(a$path, c(1, 1, 2))

    a <- search_node(wheat, `$type` = "Models.PMF.Cultivar, Models")
    expect_equal(length(a), 2)
    # Search all nodes
    a <- search_node(wheat, Name = "Simulations1", all = TRUE)
    expect_equal(length(a), 0)
    expect_equal(a, list())

    a <- search_node(wheat, Name = "Simulations", all = TRUE)
    expect_equal(length(a), 1)
    expect_equal(a[[1]]$path, 1)

    a <- search_node(wheat, Name = "Wheat", all = TRUE)
    expect_equal(length(a), 1)
    expect_equal(a[[1]]$path, c(1, 1))

    a <- search_node(wheat, Name = "Arbitrator", all = TRUE)
    expect_equal(length(a), 1)
    expect_equal(a[[1]]$path, c(1, 1, 2))
    a <- search_node(wheat, `$type` = "Models.PMF.Cultivar, Models", all = TRUE)
    expect_equal(length(a), 5)

    # Search by path
    # Absolute path
    # Root path
    a <- search_path(wheat, '.Simulations1')
    expect_equal(length(a), 0)
    expect_equal(a, list())

    a <- search_path(wheat, '.simulations', case_insensitive = FALSE)
    expect_equal(length(a), 0)
    expect_equal(a, list())

    a <- search_path(wheat, '.Simulations')
    expect_equal(length(a), 2)
    expect_equal(a$path, 1)

    a <- search_path(wheat, '.simulations')
    expect_equal(length(a), 2)
    expect_equal(a$path, 1)


    # Level one
    a <- search_path(wheat, '.Simulations.Wheat1')
    expect_equal(length(a), 0)
    expect_equal(a, list())

    # Level two
    a <- search_path(wheat, '.Simulations.Wheat')
    expect_equal(length(a), 2)
    expect_equal(a$path, c(1, 1))

    # Level three
    a <- search_path(wheat, '.Simulations.Wheat.BranchingRate')
    expect_equal(length(a), 0)
    expect_equal(a, list())

    a <- search_path(wheat, '.Simulations.Wheat.Structure')
    expect_equal(length(a), 2)
    expect_equal(a$path, c(1, 1, 4))
    # Level four
    a <- search_path(wheat, '.Simulations.Wheat.Structure.BranchingRate')
    expect_equal(length(a), 2)
    expect_equal(a$path, c(1, 1, 4, 1))
    a <- search_path(wheat, '.Simulations.Wheat.Structure.BranchingRate1')
    expect_equal(length(a), 0)
    expect_equal(a, list())

    #scoped
    # Root path
    a <- search_path(wheat, '[Simulations1]')
    expect_equal(length(a), 0)
    expect_equal(a, list())
    a <- search_path(wheat, '[Simulations]')
    expect_equal(length(a), 2)
    expect_equal(a$path, 1)

    # Level two
    a <- search_path(wheat, '[Simulations].Wheat1')
    expect_equal(length(a), 0)
    expect_equal(a, list())

    a <- search_path(wheat, '[Simulations1].Wheat')
    expect_equal(length(a), 0)
    expect_equal(a, list())

    a <- search_path(wheat, '[Whea]')
    expect_equal(length(a), 0)
    expect_equal(a, list())

    a <- search_path(wheat, '[Wheat]')
    expect_equal(length(a), 2)
    expect_equal(a$path, c(1, 1))

    # Level three
    a <- search_path(wheat, '[Wheat].BranchingRate')
    expect_equal(length(a), 0)
    expect_equal(a, list())

    a <- search_path(wheat, '[Wheat].Structure')
    expect_equal(length(a), 2)
    expect_equal(a$path, c(1, 1, 4))

    a <- search_path(wheat, '[Structure]')
    expect_equal(length(a), 2)
    expect_equal(a$path, c(1, 1, 4))

    # Level four
    a <- search_path(wheat, '[Structure].BranchingRate')
    expect_equal(length(a), 2)
    expect_equal(a$path, c(1, 1, 4, 1))

    a <- search_path(wheat, '[Structure].BranchingRate1')
    expect_equal(length(a), 0)
    expect_equal(a, list())

    a <- search_path(wheat, '[Structure1].BranchingRate')
    expect_equal(length(a), 0)
    expect_equal(a, list())

})

suppressPackageStartupMessages(library(testthat))
test_that("Walking json tree", {
    # Read wheat.json
    wheat <- read_apsimx(system.file("extdata/Wheat.json", package = "rapsimng"))

    a <- search_node(wheat, Name = "Simulations")
    expect_error(get_parent(wheat, 1))

    # Level four
    a <- search_path(wheat, '[Structure].BranchingRate')
    b <- get_parent(wheat, a$path)
    expect_equal(b$node$Name, "Structure")
    expect_equal(b$path, c(1, 1, 4))
})

get_node <- function(l) {
    if (methods::is(l, "apsimxNode")) {
        l <- l$node
    }
    if (!methods::is(l, "list")) {
        stop("A list object is required.")
    }
    l
}

#'
#' #' Convert a model into xypair
#' #'
#' #' @param l the list of apsimx file
#' #' @param lower The lower bound. The first value will be copied lower is not NULL
#' #' @param upper The upper bound. The last value will be copied upper is not NULL
#' #' @return a data.frame for X and Y values.
#' #'
#' #' @export
#' #' @examples
#' #' file <- system.file("extdata/wheat.apsimx", package = "rapsimng")
#' #' m <- read_apsimx(file)
#' #' potential <- search_path(m,
#' #'    path = '[Structure].BranchingRate.PotentialBranchingRate.Vegetative.PotentialBranchingRate')
#' #' xypair(potential)
#' xypair <- function(l, lower = NULL, upper = NULL) {
#'
#'     l_children <- l$node$Children
#'     if (is.null(l_children)) {
#'         stop('Children node is not found.')
#'     }
#'     x <- NULL
#'     Y <- NULL
#'     for (i in seq_along(l_children)) {
#'         if (!is.null(l_children[[i]]$`$type`) &&
#'             l_children[[i]]$`$type` == 'Models.Functions.XYPairs, Models') {
#'             X <- unlist(l_children[[i]]$X)
#'             Y <- unlist(l_children[[i]]$Y)
#'             break
#'         }
#'     }
#'     if (length(X) == 0 || length(Y) == 0) {
#'         stop('No values are found for X and Y.')
#'     }
#'     if (!is.null(lower)) {
#'         X <- c(lower, X)
#'         Y <- c(Y[1], Y)
#'     }
#'     if (!is.null(upper)) {
#'         X <- c(X, upper)
#'         Y <- c(Y, Y[length(Y)])
#'     }
#'     res <- data.frame(X = X, Y = Y)
#'     return(res)
#' }
#'
#'
#' #' The fixed value in the apsimx
#' #'
#' #' @param l the list of apsimx file
#' #' @export
#' #'
#' #' @return Value of this node
#' fixed_value <- function(l) {
#'     as.numeric(l$FixedValue)
#' }

---
title: "Generate apsimx file for sensitivity analysis"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Generate apsimx file for sensitivity analysis}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

```{r setup, message=FALSE, warning=FALSE}
library(rapsimng)
library(tidyverse)
```

Sensitivity analysis is conducted to check output variations when a parameter is changed. This post is to show how to conduct sensitivity analysis using rapsimng package using factorial simulations.


We use the base phyllochron as an example to demonstrate how to generate a new apsimx file from a template. 


The base phyllochron is a key parameter for wheat phelonogy and leaf appearance rate. The range of base phyllochron is from 60 to 130 degree days. 

The data.frame requires three columns (i.e. parameter, value, name) and multiple parameters can be specified here. 

```{r define-parameter}
phyllochron_para <- tibble(parameter = "[Phenology].Phyllochron.BasePhyllochron.FixedValue", 
                           value = seq(60, 130, by = 1)) %>% 
  mutate(name = paste0("Cultivar", seq_len(n())))
head(phyllochron_para)
```

The template is an apsimx file setup for the actual experiment or the specified environments (i.e. locations, sowing date or years). I assume there is a factor `Cv` for culivar in the `Permutation` model which specified the cultivar by `[Sowing].Script.CultivarName`.

```{r}
template <- read_apsimx(system.file("extdata/wheat.apsimx", package = "rapsimng"))
```


`update_cultivar` can be used to add the list of cultivars in the `Replacements`. The `Specification` in the `Permutation.Cv` can be replace with new values.

```{r}
template <- update_cultivar(template, phyllochron_para)

node <- search_path(template, "[Permutation].Cv")    
if (length(node) == 0) {
  stop("[Permutation].Cv is not found")
}
new_value <- paste0("\\1", paste(phyllochron_para$name, collapse = ","))
node$node$Specification <- gsub("(.+ *= *)(.+)$", new_value, node$node$Specification)

node$node$Specification

template <- replace_model(template, node$path, node$node)
```

Finally the new model can be write into file system and run with APSIM NG. Uncomment the sections below, update the path to `Models.exe`. 
```{r}

# write_apsimx(template, "new-path.apsimx")
# models_path <- "path-to-Models.exe"
# run_models(models_path, sim_name, csv = TRUE, parallel = FALSE)
```

The example in this post can be modified into parallel codes for sensivity analysis in the large scales. 

---
title: "Modify apsimx file"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Modify apsimx file}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE, warning=FALSE, error=FALSE, message=FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

```{r setup, echo=FALSE, warning=FALSE, error=FALSE, message=FALSE}
library(rapsimng)
suppressPackageStartupMessages(library(magrittr))
```

Several functions are designed to modify the apsimx file, including
* `replace` an existing model
* `remove` an existing model
* `append` a new model as sibling node
* `insert` a new model as child



## Read apsimx file

`read_apsimx` is used to read files with apsimx json format, i.e. `*.apsimx` for simulations and `*.json` for model under resource.

```{r read-apsimx}
library(rapsimng)
wheat <- read_apsimx(system.file("extdata/Wheat.json", package = "rapsimng"))
#wheat <- read_apsimx("inst/Wheat.json")
```

## Modify existing model
The existing model can be `replaced` and `removed`. The example below shows to update the critical thermal and then remove it.


```{r}
# Find the ThermalTime model 
a <- search_path(wheat, '[Wheat].Phenology.ThermalTime')
a$node$Children[[1]]$X[[2]]
# Update the optimum temperature
a$node$Children[[1]]$X[[2]] <- 27
# Replace with new value
wheat_new <- replace_model(wheat, a$path, a$node)
b <- search_path(wheat_new, '[Wheat].Phenology.ThermalTime')
# The optimum temperature should be updated now
b$node$Children[[1]]$X[[2]]

# The ThermalTime model can also be removed
a <- search_path(wheat, '[Wheat].Phenology.ThermalTime')
wheat_new <- remove_model(wheat, a$path)
b <- search_path(wheat_new, '[Wheat].Phenology.ThermalTime')
# The ThermalTime model should not be found now (i.e. Empty list)
b
```


## Create new model

Function `new_model` is used to generate the required elements for a new model, e.g. createing a new cultivar `Hartog`

```{r}
new_cultivar <- new_model("PMF.Cultivar", name = "Hartog")
new_cultivar
```

Then the `Command` can be updated to specify new parameter values.

```{r}
new_cultivar$Command <- list(
  "[Phenology].MinimumLeafNumber.FixedValue = 6",
  "[Phenology].VrnSensitivity.FixedValue = 0")
```

Finally the new cultivar can be inserted into apsimx file.

```{r}
# Read the apsimx file
wheat <- read_apsimx(system.file("extdata/wheat.apsimx", package = "rapsimng"))
# Create a new Replacements
replacements <- new_model("Core.Replacements")
# Insert the replacements into root folder
wheat_new <- insert_model(wheat, 1, replacements)
replacements_node <- search_path(wheat_new, ".Simulations.Replacements")
replacements_node$path

# Insert the new cultivar
wheat_new <- insert_model(wheat_new, replacements_node$path, new_cultivar)

# Check the new cultivar
cultivar_node <- search_path(wheat_new,
                                        ".Simulations.Replacements.Hartog")
cultivar_node$path
cultivar_node$node$Command
    
```


## Update cultivar parameter values

Function `update_cultivar` is a short way to update parameter values for cultivars.

```{r}
wheat <- read_apsimx(system.file("extdata/wheat.apsimx", package = "rapsimng"))
# Update cultivars
df <- data.frame(name = rep("Hartog", 3),
                  parameter = c("[Phenology].MinimumLeafNumber.FixedValue",
                               "[Phenology].VrnSensitivity.FixedValue",
                               "[Phenology].PpSensitivity.FixedValue"),
                 value = c(9, 7, 3))
wheat_cultivar <- update_cultivar(wheat, df)
# Check update cultivar paramters
hartog <- search_path(wheat_cultivar, "[Replacements].Hartog")
hartog$path
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/soil.R
\name{get_pawc}
\alias{get_pawc}
\title{Estimate Plant Available Water Capacity (PAWC) from an apsimx file}
\usage{
get_pawc(l, soil_path = NULL, crop = NULL, depth = NULL)
}
\arguments{
\item{l}{The list of apsimx file}

\item{soil_path}{The path to the soil model. If NULL (default), searches for the
first Soil model. Can be a character string specifying the soil name.}

\item{crop}{The crop name for crop-specific lower limit (LL).
If NULL (default), uses LL15 and does not apply XF. If provided, uses the
crop-specific LL and XF values from the SoilCrop parameters.}

\item{depth}{The depth (in mm) to which PAWC should be calculated.
If NULL (default), calculates for all layers. If specified, applies proportional
weighting to layers that are partially or completely beyond this depth.}
}
\value{
A numeric value of PAWC in mm
}
\description{
This function calculates the Plant Available Water Capacity (PAWC) from soil
properties in an APSIM Next Generation (.apsimx) file. PAWC represents the
amount of water that can be stored in the soil and is available for plant uptake.
}
\details{
PAWC represents the soil's water storage capacity between field capacity (DUL) and
the permanent wilting point (LL). It is a critical parameter for:

PAWC is calculated for each soil layer using the formula:

\strong{When crop = NULL and depth = NULL:}
\deqn{PAWC_i = (DUL_i - LL15_i) \times Thickness_i}

\strong{When crop is specified:}
\deqn{PAWC_i = (DUL_i - LL_i) \times Thickness_i \times XF_i}

\strong{When depth is specified (with or without crop):}
\deqn{PAWC_i = (DUL_i - LL_i) \times Thickness_i \times [XF_i] \times Weight_i}

Where:
\itemize{
\item \strong{DUL} (Drained Upper Limit): The volumetric water content (mm/mm)
after the soil has drained under gravity, typically 2-3 days after saturation
\item \strong{LL} (Lower Limit): The volumetric water content (mm/mm) below which
plants cannot extract water. This can be either:
\itemize{
\item \strong{LL15}: Generic lower limit at 15 bar suction (used when crop = NULL)
\item \strong{Crop-specific LL}: Lower limit specific to a crop type,
accounting for different rooting characteristics
}
\item \strong{Thickness}: Layer thickness in millimeters
\item \strong{XF}: Exploration factor (0-1), representing the fraction of a layer
explored by roots. Only applied when crop is specified; represents the
proportion of the layer volume that roots can explore
\item \strong{Weight}: Depth weighting factor (0-1), only applied when depth is specified:
\itemize{
\item 1.0 for layers completely within the specified depth
\item Proportional fraction for layers partially within the depth
\item 0.0 for layers completely beyond the specified depth
}
}

The total PAWC is the sum across all layers:
\deqn{PAWC_{total} = \sum_{i=1}^{n} PAWC_i}
}
\examples{
wheat <- read_apsimx(system.file("extdata/wheat.apsimx", package = "rapsimng"))

# Get total PAWC using LL15 (generic)
pawc <- get_pawc(wheat)
pawc

# Get PAWC for wheat crop (uses crop-specific LL and XF)
pawc_wheat <- get_pawc(wheat, crop = "Wheat")
pawc_wheat

# Get PAWC to 1200 mm depth
pawc_1200 <- get_pawc(wheat, depth = 1200)
pawc_1200

# Get wheat PAWC to 1500 mm depth
pawc_wheat_1500 <- get_pawc(wheat, crop = "Wheat", depth = 1500)
pawc_wheat_1500
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/modify.R
\name{set_parameter_value}
\alias{set_parameter_value}
\title{Set a parameter with a new value}
\usage{
set_parameter_value(l, parameter, value)
}
\arguments{
\item{l}{the list of apsimx file}

\item{parameter}{the name of parameter with APSIM NG specification}

\item{value}{the new value}
}
\value{
A list with replaced value
}
\description{
Set a parameter with a new value
}
\examples{
 wheat <- read_apsimx(system.file("extdata/Wheat.json", package = "rapsimng"))
 new_wheat <- set_parameter_value(wheat,
  "[Structure].BranchingRate.PotentialBranchingRate.Reproductive.Zero.FixedValue",
  1)
new_wheat2 <- search_path(new_wheat,
    "[Structure].BranchingRate.PotentialBranchingRate.Reproductive.Zero")
new_wheat2$node$FixedValue

new_wheat <- set_parameter_value(
    wheat,
    "[Structure].HeightModel.WaterStress.XYPairs.Y",
    "0.1,1.1")
new_wheat2 <- search_path(new_wheat,
    "[Structure].HeightModel.WaterStress.XYPairs")
new_wheat2$node$Y
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cultivar.R
\name{update_cultivar}
\alias{update_cultivar}
\title{Title Update the cultivar parameters}
\usage{
update_cultivar(
  l,
  df,
  add = TRUE,
  use_folder = TRUE,
  cultivar_folder = "Cultivars"
)
}
\arguments{
\item{l}{The list of apsimx file}

\item{df}{A data frame for new parameters with three columns, i.e. name, parameter and value.}

\item{add}{Whether to add extra nodes (e.g. replacements, Cultivars folder
and new cultivar)}

\item{use_folder}{use cultivar folder to add new cultivars}

\item{cultivar_folder}{folder name for cultivars}
}
\value{
The modified apsimx file
}
\description{
This function assumes the file is apsimx format.
A new Replacements node is added if it is not exist.
The existing cultivar parameters are updated. New cultivar
is created.
}
\examples{
wheat <- read_apsimx(system.file("extdata/wheat.apsimx", package = "rapsimng"))
# Update cultivars
df <- data.frame(name = rep("Hartog", 3),
                 parameter = c("[Phenology].MinimumLeafNumber.FixedValue",
                              "[Phenology].VrnSensitivity.FixedValue",
                              "[Phenology].PpSensitivity.FixedValue"),
                value = c(9, 7, 3))

wheat_cultivar <- update_cultivar(wheat, df)
hartog <- search_path(wheat_cultivar, "[Replacements].Hartog")
hartog$path
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/simulations.R
\name{keep_simulations}
\alias{keep_simulations}
\title{Keep simulations for a factorial experiment}
\usage{
keep_simulations(l, s)
}
\arguments{
\item{l}{A list from read_apsim with Factorial.Permutation as root.}

\item{s}{a list with factor as name and levels as value to keep.
The factor is kept if it is not specified.}
}
\value{
A new list with removed simulations.
}
\description{
Keep simulations for a factorial experiment
}
\examples{
wheat <- read_apsimx(system.file("extdata/wheat.apsimx", package = "rapsimng"))
permutation <- search_path(wheat, path = "[Factors].Permutation")
permutation_new <- keep_simulations(permutation$node, list(V = "2"))
get_simulations(permutation_new)

permutation_new <- keep_simulations(permutation$node, list(Cv = c("Axe", "Bolac")))
get_simulations(permutation_new)

permutation_new <- keep_simulations(permutation$node,
                                    list(V = "1", Cv = c("Axe", "Bolac")))
get_simulations(permutation_new)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/run_models.R
\name{run_models}
\alias{run_models}
\title{Run apsimx file using Models.exe}
\usage{
run_models(
  models_exe,
  path,
  pattern = NULL,
  recurse = FALSE,
  csv = FALSE,
  parallel = NULL,
  ncpus = NULL,
  verbose = FALSE
)
}
\arguments{
\item{models_exe}{path to Models.exe}

\item{path}{The path to an .apsimx file. May include wildcard.}

\item{pattern}{Use to filter simulation names to run.}

\item{recurse}{Recursively search subdirectories for files matching ApsimXFileSpec. FALSE in default.}

\item{csv}{Export all reports to .csv files. FALSE in default.}

\item{parallel}{Use the multi-process job runner. If FALSE, use single threaded; if TRUE, use the multi-process job runner}

\item{ncpus}{Set the number of processors to use. All processes in default}

\item{verbose}{Write messages to StdOut when a simulation starts/finishes. Only has an effect when running a directory of .apsimx files (*.apsimx).}
}
\description{
Run apsimx file using Models.exe
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/search.R
\name{search_node}
\alias{search_node}
\title{Find element(s) in  apsimx file}
\usage{
search_node(l, all = FALSE, max_depth = 1e+06, case_insensitive = TRUE, ...)
}
\arguments{
\item{l}{The list of apsimx file}

\item{all}{Whether to find all elements}

\item{max_depth}{The maximum depth to search}

\item{case_insensitive}{Whether case sensitive}

\item{...}{Other names arguments for property to match}
}
\value{
A list matching all criteria if all equals to TRUE,
A list with node and path if all equals to FALSE (default)
}
\description{
Find element(s) in  apsimx file
}
\examples{
wheat <- read_apsimx(system.file("extdata/Wheat.json", package = "rapsimng"))
# Return empty list if not found
search_node(wheat, Name = "Simulations1")
# Find root level
a <- search_node(wheat, Name = "Simulations")
a$path
# Find sub-level
a <- search_node(wheat, Name = "Wheat")
a$path
a <- search_node(wheat, `$type` = "Models.PMF.Cultivar, Models")
a$path

# Find multiple attributes
a <- search_node(wheat,
            Name = 'PotentialBranchingRate',
            `$type` = "Models.Functions.PhaseLookup, Models")
a$path
a$node$Name
# Find all cultivar nodes
 a <- search_node(wheat, `$type` = "Models.PMF.Cultivar, Models", all = TRUE)
length(a)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/simulations.R
\name{get_simulations}
\alias{get_simulations}
\title{Get simulations for a factorial experiment}
\usage{
get_simulations(l)
}
\arguments{
\item{l}{A list from read_apsim with Factorial.Permutation as root.}
}
\value{
A list with Factor as name and Levels as values
}
\description{
Get simulations for a factorial experiment
}
\examples{
wheat <- read_apsimx(system.file("extdata/wheat.apsimx", package = "rapsimng"))
permutation <- search_path(wheat, path = "[Factors].Permutation")
get_simulations(permutation$node)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cultivar.R
\name{new_cultivar}
\alias{new_cultivar}
\title{Generate new cultivars with parameter which can be used in Replacements}
\usage{
new_cultivar(df, cultivar_folder = "Cultivars")
}
\arguments{
\item{df}{A data frame for new parameters with three columns, i.e. name, parameter and value.}

\item{cultivar_folder}{folder name for cultivars}
}
\value{
An APSIMX list
}
\description{
Generate new cultivars with parameter which can be used in Replacements
}
\examples{
df <- data.frame(name = c("C1", "C1", "C2", "C2"),
     parameter = c("[Phenology].CAMP.FLNparams.MinLN",
              "[Phenology].CAMP.FLNparams.VrnLN",
              "[Phenology].CAMP.FLNparams.MinLN",
              "[Phenology].CAMP.FLNparams.VrnLN"),
     value = c(5, 6, 7, 8))
a <- new_cultivar(df)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/modify.R
\name{append_model}
\alias{append_model}
\title{append a model into apsimx}
\usage{
append_model(l, path, model)
}
\arguments{
\item{l}{the list of apsimx file}

\item{path}{If numeric, the path returned by search_path or search_node. If character, the path supported by apsimx}

\item{model}{A new model which should be a list of new models}
}
\value{
The modified list with new value
}
\description{
append a model into apsimx
}
\examples{
wheat <- read_apsimx(system.file("extdata/wheat.apsimx", package = "rapsimng"))
replacements <- new_model("Core.Replacements")
wheat_new <- insert_model(wheat, 1, replacements)
replacements_node <- search_path(wheat_new, ".Simulations.Replacements")
replacements_node$path
# Add a cultivar folder under replacements
cultivar_folder <- new_model("PMF.CultivarFolder", "Cultivars")
wheat_new <- insert_model(wheat_new, replacements_node$path, cultivar_folder)
cultivar_folder_node <- search_path(wheat_new,
                                    ".Simulations.Replacements.Cultivars")
cultivar_folder_node$path
# Add an new cultivar
cultivar <- new_model("PMF.Cultivar", "Hartog")
wheat_new <- insert_model(wheat_new, cultivar_folder_node$path, cultivar)
cultivar_node <- search_path(wheat_new,
                             ".Simulations.Replacements.Cultivars.Hartog")
cultivar_node$path
# Append another cultivar
cultivar2 <- new_model("PMF.Cultivar", "Axe")
wheat_new <- append_model(wheat_new, cultivar_node$path, list(cultivar2))
cultivar2_node <- search_path(wheat_new,
                              ".Simulations.Replacements.Cultivars.Axe")
cultivar2_node$path
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/modify.R
\name{remove_model}
\alias{remove_model}
\title{Remove a model with new values}
\usage{
remove_model(l, path)
}
\arguments{
\item{l}{the list of apsimx file}

\item{path}{If numeric, the path returned by search_path or search_node. If character, the path supported by apsimx}
}
\value{
The modified list with new value
}
\description{
Remove a model with new values
}
\examples{
wheat <- read_apsimx(system.file("extdata/Wheat.json", package = "rapsimng"))
a <- search_path(wheat, '[Wheat].Phenology.ThermalTime')
wheat_new <- remove_model(wheat, a$path)
b <- search_path(wheat_new, '[Wheat].Phenology.ThermalTime')
b
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/models.R
\name{new_model}
\alias{new_model}
\title{Create a new model}
\usage{
new_model(model, name = model)
}
\arguments{
\item{model}{The name of new model}

\item{name}{The new name}
}
\description{
Create a new model
}
\examples{
new_model(model = "PMF.Cultivar")
new_model(model = "PMF.Cultivar", name = "example")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cultivar.R
\name{get_cultivar}
\alias{get_cultivar}
\title{Get all cultivar parameters in a model}
\usage{
get_cultivar(l, alias = TRUE)
}
\arguments{
\item{l}{The list of apsimx file}

\item{alias}{Whether to export alias}
}
\value{
A data frame for all cultivar parameters
}
\description{
Get all cultivar parameters in a model
}
\examples{
wheat <- read_apsimx(system.file("extdata/Wheat.json", package = "rapsimng"))
get_cultivar(wheat)
get_cultivar(wheat, alias = FALSE)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/report.R
\name{list_report}
\alias{list_report}
\title{List all reports in the database}
\usage{
list_report(file)
}
\arguments{
\item{file}{the file path to apsimx or db file}
}
\value{
a vector of all reports
}
\description{
List all reports in the database
}
\examples{
\dontrun{
file <- system.file("extdata/wheat.apsimx", package = "rapsimng")
list_report(file)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/modify.R
\name{disable_models}
\alias{disable_models}
\title{Disable models in apsimx}
\usage{
disable_models(l, paths)
}
\arguments{
\item{l}{the list of apsimx file}

\item{paths}{If numeric, the path returned by search_path or search_node. If character, the path supported by apsimx}
}
\value{
The modified list with new value
}
\description{
Disable models in apsimx
}
\examples{
wheat <- read_apsimx(system.file("extdata/Wheat.json", package = "rapsimng"))
a <- disable_models(wheat, '[Wheat].Phenology.ThermalTime')
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/apsimx_build.R
\name{install_apsimng}
\alias{install_apsimng}
\title{Install ApsimNG Software}
\usage{
install_apsimng(
  repo = "https://github.com/APSIMInitiative/ApsimX.git",
  branch = "master",
  install_dir = "apsimx_build",
  overwrite = FALSE
)
}
\arguments{
\item{repo}{Character string specifying the repository of ApsimNG to install.}

\item{branch}{Character string specifying the branch of the repository to install.}

\item{install_dir}{Character string specifying the directory where ApsimNG
should be installed.}

\item{overwrite}{Logical. If TRUE, overwrites any existing installation
at the specified install directory.}
}
\value{
No return value.
}
\description{
This function installs the ApsimNG simulation software on the local system.
It handles the download and installation process for the APSIM Next Generation
agricultural modeling platform.
}
\details{
This function downloads and installs ApsimNG from the official repository.
It performs version checking, handles dependencies, and configures the
installation for use with R. The function requires administrative privileges
on Windows systems.
}
\examples{
\dontrun{
# Install latest version
install_apsimng()
install_apsimng(install_dir = "C:/MyPrograms/ApsimNG")
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/program.R
\name{minimum_apsimng}
\alias{minimum_apsimng}
\title{Create the minimum requirements to run an APSIM Next Generation}
\usage{
minimum_apsimng(install_path, output)
}
\arguments{
\item{install_path}{The installed path of APSIM Next Generation}

\item{output}{The output folder}
}
\description{
Create the minimum requirements to run an APSIM Next Generation
}
\examples{
\dontrun{
 minimum_apsimng("C:/ProgramFiles/APSIMNG", "minimum_apsimng")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/run_models.R
\name{with_apsimx}
\alias{with_apsimx}
\title{Run APSIM Next Generation Simulations in a Temporary Folder}
\usage{
with_apsimx(
  models,
  file,
  mets = NULL,
  target = tempdir(),
  clean = c("none", "simulations", "mets", "all"),
  post_process = NULL,
  ...
)
}
\arguments{
\item{models}{A character string specifying the path to the APSIM NG executable (\code{Models.exe} on Windows
or \code{Models} on Linux).}

\item{file}{A character string specifying the path to the \code{.apsimx} simulation file.}

\item{mets}{A character vector specifying paths to meteorological data (\code{.met} files) used in the simulation.
Currently, only \code{.met} files located in the same folder as the \code{.apsimx} file are supported (optional).}

\item{target}{A character string specifying the target directory where simulations will be run.
Defaults to the R system temporary directory (\code{tempdir()}).}

\item{clean}{A character string specifying which files or directories to clean after the simulation. Options are:
\itemize{
\item \code{"none"}: No files are cleaned (default).
\item \code{"simulations"}: Cleans only simulation-related files (e.g., \verb{*.apsimx}, \verb{*.db}, \verb{*.db-shm}, \verb{*.db-wal}, \verb{*.csv}).
\item \code{"mets"}: Cleans only meteorological files (e.g., \verb{*.met}).
\item \code{"all"}: Cleans all files, including APSIM NG executable files, simulations, and meteorological files.
}}

\item{post_process}{An optional function for post-processing simulation results.
The function must accept a \code{folder} argument specifying the directory
containing the simulation results.}

\item{...}{Additional arguments passed to both the \code{run_models} and \code{post_process} functions.}
}
\value{
If a \code{post_process} function is provided, its return value is returned. Otherwise, the function returns \code{NULL}.
}
\description{
The \code{with_apsimx} function automates the process of setting up and running APSIM Next Generation (NG) simulations
in a temporary or specified folder. This approach helps reduce I/O overhead, particularly in high-performance
computing (HPC) environments with network file systems (NFS). The frequent file I/O operations performed by APSIM NG
(e.g., writing to SQLite databases) can strain file systems, and redirecting simulations to a local folder mitigates this issue.
The function can also perform optional post-processing of the simulation results.
}
\examples{
\dontrun{
# Run simulations without post-processing
with_apsimx(
    models = "path/to/apsimx",
    file = "path/to/input.apsimx",
    mets = c("path/to/met1.met", "path/to/met2.met"),
    clean = "all"
)

# Run simulations with post-processing
post_process_function <- function(folder) {
    output_files <- list.files(folder, full.names = TRUE)
    message("Output files:", paste(output_files, collapse = "\n"))
}

result <- with_apsimx(
    models = "path/to/apsimx",
    file = "path/to/input.apsimx",
    mets = c("path/to/met1.met", "path/to/met2.met"),
    post_process = post_process_function
)
}

}
\seealso{
\code{\link{run_models}}: Runs APSIM simulations.
\code{\link{do.call}}: Dynamically calls functions with named arguments.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/experiment.R
\name{get_experiment}
\alias{get_experiment}
\title{Get an experiment node by name}
\usage{
get_experiment(apsimx, experiment)
}
\arguments{
\item{apsimx}{A parsed APSIMX model from \code{read_apsimx()}.}

\item{experiment}{A single experiment name to locate.}
}
\value{
An \code{apsimxNode} object for the experiment.
}
\description{
Get an experiment node by name
}
\examples{
wheat <- read_apsimx(system.file("extdata/wheat.apsimx", package = "rapsimng"))
exp <- search_path(wheat, path = "[Experiment]")
get_experiment(wheat, exp$node$Name)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/apsimx_build.R
\name{test_apsimx}
\alias{test_apsimx}
\title{Test whether all files under published folder of apsimx are required}
\usage{
test_apsimx(base, example)
}
\arguments{
\item{base}{the base folder path to apsimx publish}

\item{example}{an example apsimx file}
}
\value{
A vector a required files
}
\description{
Test whether all files under published folder of apsimx are required
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/simulations.R
\name{get_metfile}
\alias{get_metfile}
\title{Get the met file name for an experiment}
\usage{
get_metfile(l, is_stop = TRUE)
}
\arguments{
\item{l}{A list or apsimxNode red by read_apsimx}

\item{is_stop}{Whether stop the function when error}
}
\value{
The met file name in a experiment
}
\description{
Get the met file name for an experiment
}
\examples{
wheat <- read_apsimx(system.file("extdata/wheat.apsimx", package = "rapsimng"))
exp <- search_path(wheat, path = "[Experiment]")
get_metfile(exp)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/report.R
\name{read_report}
\alias{read_report}
\title{Read apsimx database in db file format}
\usage{
read_report(file, report)
}
\arguments{
\item{file}{the file path to apsimx or db file}

\item{report}{the report name}
}
\value{
a data.frame for a report
}
\description{
Read apsimx database in db file format
}
\examples{
\dontrun{
file <- system.file("extdata/wheat.apsimx", package = "rapsimng")
read_report(file, "HarvestReport")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_frost_heat_damage.R
\name{frost_heat_damage}
\alias{frost_heat_damage}
\alias{default_frost_heat_params}
\alias{stage_sens_frost}
\alias{stage_sens_heat}
\alias{temp_damage_frost}
\alias{temp_damage_heat}
\alias{daily_damage_frost}
\alias{daily_damage_heat}
\alias{daily_damage_frost_heat}
\alias{cum_damage_frost}
\alias{cum_damage_heat}
\alias{cum_damage_frost_heat}
\alias{calc_daily_damage_frost_heat}
\alias{calc_cum_damage_frost_heat}
\title{Frost and Heat Damage Functions}
\usage{
default_frost_heat_params()

stage_sens_frost(growth_stage, params)

stage_sens_heat(growth_stage, params)

temp_damage_frost(mint, params)

temp_damage_heat(maxt, params)

daily_damage_frost(mint, growth_stage, params)

daily_damage_heat(maxt, growth_stage, params)

daily_damage_frost_heat(mint, maxt, growth_stage, params)

cum_damage_frost(mint, growth_stage, params)

cum_damage_heat(maxt, growth_stage, params)

cum_damage_frost_heat(mint, maxt, growth_stage, params)

calc_daily_damage_frost_heat(
  mint,
  maxt,
  growth_stage,
  crop = c("Wheat", "Canola"),
  params = NULL
)

calc_cum_damage_frost_heat(
  mint,
  maxt,
  growth_stage,
  crop = c("Wheat", "Canola"),
  params = NULL
)
}
\arguments{
\item{growth_stage}{Numeric vector of daily growth stage.}

\item{params}{A crop parameter list. If provided, it will override the \code{crop} argument.}

\item{mint}{Numeric vector of daily minimum temperature.}

\item{maxt}{Numeric vector of daily maximum temperature.}

\item{crop}{Crop name to use from \code{default_frost_heat_params()}.}
}
\value{
A named list with crop-specific parameter lists.

A numeric vector of frost stage sensitivity (0-1).

A numeric vector of heat stage sensitivity (0-1).

A numeric vector of frost damage contribution from temperature.

A numeric vector of heat damage contribution from temperature.

A numeric vector of daily frost damage.

A numeric vector of daily heat damage.

A numeric vector of combined daily frost and heat damage.

A scalar cumulative frost damage.

A scalar cumulative heat damage.

A scalar cumulative frost and heat damage.

A tibble with daily stage sensitivity, temperature damage, and daily
frost/heat/combined damage.

A tibble with cumulative frost damage, cumulative heat damage, and
cumulative combined frost-heat damage.
}
\description{
Helper and wrapper functions to calculate frost and heat damage from
temperature and growth stage.
}
\details{
The damage functions are based on the APSIM NG Model \href{https://github.com/APSIMInitiative/ApsimX/blob/3787815684da28c13467942b9d886fa04f54d4c6/Models/Functions/FrostHeatDamageFunctions.cs}{FrostHeatDamageFunctions}. See full documentation \href{https://notes.apsimng.bangyou.me/docs/Models/Functions/FrostHeatDamageFunctions.html}{here}.
}
\examples{
default_frost_heat_params()
calc_daily_damage_frost_heat(
  mint = c(-2, 0, 4),
  maxt = c(29, 33, 36),
  growth_stage = c(7.0, 8.0, 9.0),
  crop = "Wheat"
)
calc_cum_damage_frost_heat(
  mint = c(-2, 0, 4),
  maxt = c(29, 33, 36),
  growth_stage = c(7.0, 8.0, 9.0),
  crop = "Wheat"
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/search.R
\name{search_path}
\alias{search_path}
\title{Find a model in the apsimx file using specified path}
\usage{
search_path(l, path, case_insensitive = TRUE)
}
\arguments{
\item{l}{the list of apsimx file}

\item{path}{The specified path (See details)}

\item{case_insensitive}{Whether case sensitive}
}
\value{
The list for the specified path.
}
\description{
Find a model in the apsimx file using specified path
}
\section{Absolute Paths}{

Absolute paths have a leading ‘.’ e.g.
\itemize{
\item .Simulations.Test.Clock - absolute path - refers to the clock model
in the 'Test' simulation.
}
}

\section{Scoped Paths}{

Scoped paths have a leading model type in square brackets. A model
of the specified name, in scope, is located before applying the rest of the path.
\itemize{
\item [Soil].Water - scoped path - refers to the Water model that is a child of a model
that has the name 'Soil' that is in scope
}
}

\examples{
wheat <- read_apsimx(system.file("extdata/Wheat.json", package = "rapsimng"))
# Return empty list if not found
search_path(wheat, "[Simulations1]")
# Search root path
a <- search_path(wheat, '.Simulations')
a$path
a$node$Name
# Level one
a <- search_path(wheat, '.Simulations.Wheat1')
a$path
a$node$Name
# Level two
a <- search_path(wheat, '.Simulations.Wheat')
a$path
a$node$Name
# Level three
a <- search_path(wheat, '.Simulations.Wheat.BranchingRate')
a$path
a$node$Name
a <- search_path(wheat, '.Simulations.Wheat.Structure')
a$path
a$node$Name
# Level four
a <- search_path(wheat, '.Simulations.Wheat.Structure.BranchingRate')
a$path
a$node$Name
a <- search_path(wheat, '.Simulations.Wheat.Structure.BranchingRate1')
a$path
a$node$Name
# scoped
# Root path
a <- search_path(wheat, '[Simulations1]')
a <- search_path(wheat, '[Simulations]')
a$path
a$node$Name
# Level two
a <- search_path(wheat, '[Simulations].Wheat1')
a <- search_path(wheat, '[Simulations1].Wheat')
a$path
a$node$Name
a <- search_path(wheat, '[Whea]')
a <- search_path(wheat, '[Wheat]')
a$path
a$node$Name
# Level three
a <- search_path(wheat, '[Wheat].BranchingRate')
a <- search_path(wheat, '[Wheat].Structure')
a$path
a$node$Name
a <- search_path(wheat, '[Structure]')
a$path
a$node$Name
# Level four
a <- search_path(wheat, '[Structure].BranchingRate')
a$path
a$node$Name
a <- search_path(wheat, '[Structure].BranchingRate1')
a <- search_path(wheat, '[Structure1].BranchingRate')
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/file.R
\name{read_apsimx}
\alias{read_apsimx}
\title{Read APSIMX file}
\usage{
read_apsimx(path)
}
\arguments{
\item{path}{The file path or URL to apsimx file}
}
\value{
A list object of apsimx file
}
\description{
Read APSIMX file
}
\examples{
file <- system.file("extdata/wheat.apsimx", package = "rapsimng")
m <- read_apsimx(file)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/json_tree.R
\name{get_parent}
\alias{get_parent}
\title{Get the parent node from a path}
\usage{
get_parent(l, path)
}
\arguments{
\item{l}{the list of apsimx file}

\item{path}{If numeric, the path returned by search_path or search_node. If character, the path supported by apsimx}
}
\value{
A new list for parent
}
\description{
Get the parent node from a path
}
\examples{
wheat <- read_apsimx(system.file("extdata/Wheat.json", package = "rapsimng"))
a <- search_path(wheat, '[Structure].BranchingRate')
get_parent(wheat, a$path)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/split_apsimx.R
\name{split_apsimx}
\alias{split_apsimx}
\title{Split an APSIMX file into separate simulations except cultivar}
\usage{
split_apsimx(file, output)
}
\arguments{
\item{file}{Character. Path to the input APSIMX file.}

\item{output}{Character. Path to the directory where the split components will be saved.}
}
\value{
No return value. The function is called for its side effects.
}
\description{
This function takes an APSIMX file and splits it into separate simulations for all factors except cultivar factors.
Then saving the results to the specified output location.
}
\examples{
\dontrun{
split_apsimx("path/to/input.apsimx", "path/to/output/directory")
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/log.R
\name{log_level}
\alias{log_level}
\title{Set the log level of apsimx file}
\usage{
log_level(l, level = c("Error", "Warning", "Information", "Diagnostic", "All"))
}
\arguments{
\item{l}{the list of apsimx file}

\item{level}{log level with option Error, Warning, Information, Diagnostic, All}
}
\value{
a new apsimx file
}
\description{
Set the log level of apsimx file
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/file.R
\name{write_apsimx}
\alias{write_apsimx}
\title{Write APSIMX file}
\usage{
write_apsimx(l, file)
}
\arguments{
\item{l}{the list of apsimx file}

\item{file}{The file path to apsimx file}
}
\value{
A list object of apsimx file
}
\description{
Write APSIMX file
}
